﻿/**
 * Редактор Crop для DocumentEditor (расширение)
 * @returns {} 
 */
batchApp.datacontext.CropEditor = function (documentEditor, cropTitle, lock, dataContext, onDeactivate, onUpdatePage) {
	var self = this;

	self.documentEditor = documentEditor;
	self.lock = lock;
	self.dataContext = dataContext;
	self.onDeactivate = onDeactivate;
	self.onUpdatePage = onUpdatePage;
	self.isEnabled = false;
	self.rectangles = new Map();
	self.changedIds = new Set();
	// Данные текущего действия
	self.activePageData = null;
	self.action = 0;

	self.documentEditor.addOnMouseDown(self.onMouseDown.bind(this));
	self.documentEditor.addOnMouseMove(self.onMouseMove.bind(this));
	self.documentEditor.addOnMouseUp(self.onMouseUp.bind(this));

	self.documentEditor.addOnMouseLeaveGlobal(function (e) {
		if (e.relatedTarget !== self.cropButton && e.target !== self.cropButton) {
			self.cropButton.style.display = 'none';
		}
	});

	$(document).keypress(self.onKeyDown.bind(this));

	self.documentEditor.onOpenedDocumentChanged(function () {
		if (self.isEnabled) {
			self.toggle();
		}
	});

	self.cropButton = document.createElement('button');
	self.cropButton.className = CROP_BUTTON_CLASS_NAME;
	self.cropButton.innerText = cropTitle;
	self.cropButton.onclick = self.crop.bind(self);
	document.getElementById('documentEditor__extraLayer').appendChild(self.cropButton);
};

var CROP_BUTTON_CLASS_NAME = 'cropEditor__crop';

batchApp.datacontext.CropEditor.LineLength = 30;
batchApp.datacontext.CropEditor.LineThickness = 2;
batchApp.datacontext.CropEditor.ActiveRange = 10;

batchApp.datacontext.CropEditor.Actions = { None: 0, Edit: 1, Move: 2 };
batchApp.datacontext.CropEditor.cropCursor = "url(" +
	FC.Environment.AppPathNoTennant +
	"content/images/crosshair_crop.cur), auto";

batchApp.datacontext.CropEditor.prototype.onKeyDown = function (event) {
	if (!this.isEnabled || event.keyCode !== 13) {
		return true;
	}

	this.crop();

	return false;
};

/**
 * Зафиксировать Crop
 * @returns {} 
 */
batchApp.datacontext.CropEditor.prototype.crop = function () {
	var self = this;

	if (self.lock()) {
		return;
	}

	this.changedIds.forEach(function (pageId) {
		var rect = self.rectangles.get(pageId);
		var page = self.documentEditor.getPageById(pageId);
		self.documentEditor.setCursor(batchApp.datacontext.CropEditor.cropCursor);

		self.dataContext.cropPage(pageId, Math.round(rect.left), Math.round(rect.top),
			Math.round(rect.left + rect.width), Math.round(rect.top + rect.height), function (result) {
				if (result.hasErrors) {
					FC.Utils.Error.Show(result.errorText);
					return;
				}

				self.onUpdatePage(pageId);
			});
	});

	self.toggle();
	this.cropButton.style.display = 'none';
};

batchApp.datacontext.CropEditor.prototype.onMouseDown = function (pageData, e) {
	if (!this.isEnabled) {
		return true;
	}

	var rect = this.rectangles.get(pageData.pageId);
	if (!rect) {
		return true;
	}

	var page = this.documentEditor.getPageById(pageData.pageId);
	if (!page) {
		return true;
	}

	var cropButton = $(this.cropButton);

	// Проверим попадание в границы текущего прямоугольника
	var hit = this.hitTestRect(rect, pageData);
	if (hit) {
		this.activePageData = {
			pageId: pageData.pageId,
			x: pageData.x,
			y: pageData.y,
			hit: hit,
			rect: rect
		};

		this.documentEditor.captureMouse(pageData.pageId);
		cropButton.hide();
		return false;
	}

	if (pageData.x < 0 || pageData.x > page.imageWidth || pageData.y < 0 || pageData.y > page.imageHeight) {
		return true;
	}

	this.activePageData = {
		pageId: pageData.pageId,
		x: pageData.x,
		y: pageData.y,
		rect: rect
	};

	var insideRect = this.isInsideRect(rect, pageData);
	var isChanged = this.changedIds.has(pageData.pageId);
	if (insideRect && isChanged) {
		// В этом случае осуществляется не рисование, а перенос прямоугольника
		this.action = batchApp.datacontext.CropEditor.Actions.Move;

		var zoom = this.documentEditor.zoom();
		this.activePageData.cropPos = {
			x: parseInt(cropButton.css("left")) / zoom - rect.left,
			y: parseInt(cropButton.css("top")) / zoom - rect.top
		};
	} else {
		this.action = batchApp.datacontext.CropEditor.Actions.Edit;
	}

	this.documentEditor.captureMouse(pageData.pageId);
	cropButton.hide();
	return false;
};

batchApp.datacontext.CropEditor.Directions = {
	TopLeft: 1,
	Top: 2,
	TopRight: 3,
	Right: 4,
	BottomRight: 5,
	Bottom: 6,
	BottomLeft: 7,
	Left: 8
};

batchApp.datacontext.CropEditor.Cursors = ['nw-resize', 'n-resize', 'ne-resize', 'e-resize', 'se-resize', 's-resize', 'sw-resize', 'w-resize'];

/**
 * Проверить попадание точки внутрь прямоугольника
 * @param {} rect 
 * @param {} point 
 * @returns {} 
 */
batchApp.datacontext.CropEditor.prototype.isInsideRect = function (rect, point) {
	var right = rect.left + rect.width;
	var bottom = rect.top + rect.height;

	return rect.left <= point.x &&
		rect.top <= point.y &&
		right >= point.x &&
		bottom >= point.y;
};

/**
 * Проверить попадание точки на границу (в допустимых пределах) прямоугольника
 * @param {} rect Прямоугольник
 * @param {} point Проверяемая точка
 * @returns {} 
 */
batchApp.datacontext.CropEditor.prototype.hitTestRect = function (rect, pageData) {
	var zoom = this.documentEditor.zoom();
	var maxDistance = batchApp.datacontext.CropEditor.ActiveRange / zoom;
	var lineLength = batchApp.datacontext.CropEditor.LineLength / zoom;

	var right = rect.left + rect.width;
	var bottom = rect.top + rect.height;

	if (rect.left - pageData.x > maxDistance ||
		rect.top - pageData.y > maxDistance ||
		pageData.x - right > maxDistance ||
		pageData.y - bottom > maxDistance) {
		return null; // Точка далеко за пределами прямоугольника
	}

	var horizontalCornerLength = Math.min(lineLength, rect.width / 3);
	var verticalCornerLength = Math.min(lineLength, rect.height / 3);

	var horizontalInnerDistance = Math.min(maxDistance, rect.width / 3);
	var verticalInnerDistance = Math.min(maxDistance, rect.height / 3);

	var yPos = pageData.y;
	var xPos = pageData.x;
	if (pageData.x - rect.left <= horizontalInnerDistance && pageData.x - rect.left >= -maxDistance) { // попадание слева
		return yPos < rect.top + verticalCornerLength
			? batchApp.datacontext.CropEditor.Directions.TopLeft
			: yPos > bottom - verticalCornerLength
			? batchApp.datacontext.CropEditor.Directions.BottomLeft
			: batchApp.datacontext.CropEditor.Directions.Left;
	}

	if (pageData.x - right <= maxDistance && pageData.x - right >= -horizontalInnerDistance) { // попадание справа
		return yPos < rect.top + verticalCornerLength
			? batchApp.datacontext.CropEditor.Directions.TopRight
			: yPos > bottom - verticalCornerLength
			? batchApp.datacontext.CropEditor.Directions.BottomRight
			: batchApp.datacontext.CropEditor.Directions.Right;
	}

	if (pageData.y - rect.top <= verticalInnerDistance && pageData.y - rect.top >= -maxDistance) { // попадание сверху
		return xPos < rect.left + horizontalCornerLength
			? batchApp.datacontext.CropEditor.Directions.TopLeft
			: xPos > right - horizontalCornerLength
			? batchApp.datacontext.CropEditor.Directions.TopRight
			: batchApp.datacontext.CropEditor.Directions.Top;
	}

	if (pageData.y - bottom <= maxDistance && pageData.y - bottom >= -verticalInnerDistance) { // попадание снизу
		return xPos < rect.left + horizontalCornerLength
			? batchApp.datacontext.CropEditor.Directions.BottomLeft
			: xPos > right - horizontalCornerLength
			? batchApp.datacontext.CropEditor.Directions.BottomRight
			: batchApp.datacontext.CropEditor.Directions.Bottom;
	}

	return null;
};

batchApp.datacontext.CropEditor.prototype.onMouseMove = function (pageData, e) {
	if (!this.isEnabled) {
		return true;
	}

	var page = this.documentEditor.getPageById(pageData.pageId);
	if (!page) {
		return true;
	}

	if (!this.activePageData) {
		// Проверить, нужно ли изменить вид курсора
		var rect = this.rectangles.get(pageData.pageId);
		if (!rect) {
			return true;
		}

		var newCursor = null;
		var hit = this.hitTestRect(rect, pageData);
		if (hit) {
			newCursor = batchApp.datacontext.CropEditor.Cursors[hit - 1];
		} else {
			newCursor = batchApp.datacontext.CropEditor.cropCursor;
		}

		var insideRect = this.isInsideRect(rect, pageData);
		var isChanged = this.changedIds.has(pageData.pageId);
		var cropButton = $(this.cropButton);
		var isButtonVisible = cropButton.is(":visible");
		if (insideRect && isChanged) {
			if (!hit) {
				newCursor = 'move';
			}

			if (!isButtonVisible && !this.lock()) {
				var zoom = this.documentEditor.zoom();

				var pageTitleHeight = 20;

				var xPos = e.offsetX;
				var yPos = e.offsetY;

				var cropButtonWidth = cropButton.outerWidth() / zoom;
				var cropButtonHeight = cropButton.outerHeight() / zoom;

				if (rect.width > cropButtonWidth && rect.height > cropButtonHeight) {
					// Кнопку можно разместить внутри выделенного прямоугольника
					
					var right = rect.left + rect.width;
					var bottom = rect.top + rect.height;

					if (right - pageData.x < pageData.x - rect.left) {
						xPos -= cropButton.outerWidth();
					}

					if (bottom - pageData.y < pageData.y - rect.top) {
						yPos -= cropButton.outerHeight();
					}
				}

				cropButton.css("left", xPos + "px");
				cropButton.css("top", yPos + "px");
				cropButton.show();
			}
		} else {
			if (isButtonVisible) {
				cropButton.hide();
			}
		}

		if (newCursor) {
			this.documentEditor.setCursor(newCursor);
		}

		return true;
	}

	if (this.activePageData.pageId !== pageData.pageId) {
		return true;
	}

	if (pageData.x < 0) {
		pageData.x = 0;
	} else if (pageData.x > page.imageWidth) {
		pageData.x = page.imageWidth;
	}

	if (pageData.y < 0) {
		pageData.y = 0;
	} else if (pageData.y > page.imageHeight) {
		pageData.y = page.imageHeight;
	}

	// Обновить прямоугольник выделения
	if (this.activePageData.hit) {
		var rect = this.activePageData.rect;
		var point1, point2; // Сформируем новый прямоугольник по двум точкам

		switch (this.activePageData.hit) {
			case batchApp.datacontext.CropEditor.Directions.TopLeft:
				point1 = pageData;
				point2 = { x: rect.left + rect.width, y: rect.top + rect.height };
				break;

			case batchApp.datacontext.CropEditor.Directions.Top:
				point1 = { x: rect.left + rect.width, y: pageData.y };
				point2 = { x: rect.left, y: rect.top + rect.height };
				break;

			case batchApp.datacontext.CropEditor.Directions.TopRight:
				point1 = pageData;
				point2 = { x: rect.left, y: rect.top + rect.height };
				break;

			case batchApp.datacontext.CropEditor.Directions.Right:
				point1 = { x: pageData.x, y: rect.top };
				point2 = { x: rect.left, y: rect.top + rect.height };
				break;

			case batchApp.datacontext.CropEditor.Directions.BottomRight:
				point1 = pageData;
				point2 = { x: rect.left, y: rect.top };
				break;

			case batchApp.datacontext.CropEditor.Directions.Bottom:
				point1 = { x: rect.left + rect.width, y: pageData.y };
				point2 = { x: rect.left, y: rect.top };
				break;

			case batchApp.datacontext.CropEditor.Directions.BottomLeft:
				point1 = pageData;
				point2 = { x: rect.left + rect.width, y: rect.top };
				break;

			case batchApp.datacontext.CropEditor.Directions.Left:
				point1 = { x: pageData.x, y: rect.top };
				point2 = { x: rect.left + rect.width, y: rect.top + rect.height };
				break;
		}

		if (point1 && point2) {
			var newRect = this.createRect(point1, point2);
			this.updateRect(page, newRect, true);
		}

		return false;
	}

	var newRect;
	if (this.action === batchApp.datacontext.CropEditor.Actions.Edit) {
		newRect = {
			left: Math.min(this.activePageData.x, pageData.x),
			top: Math.min(this.activePageData.y, pageData.y),
			width: Math.abs(this.activePageData.x - pageData.x),
			height: Math.abs(this.activePageData.y - pageData.y)
		};
	} else { // Move
		var deltaX = pageData.x - this.activePageData.x;
		var deltaY = pageData.y - this.activePageData.y;

		var newLeft = Math.max(0, Math.min(page.imageWidth - this.activePageData.rect.width, this.activePageData.rect.left + deltaX));
		var newTop = Math.max(0, Math.min(page.imageHeight - this.activePageData.rect.height, this.activePageData.rect.top + deltaY));

		newRect = {
			left: newLeft,
			top: newTop,
			width: this.activePageData.rect.width,
			height: this.activePageData.rect.height
		};

		// Кнопка Crop тоже едет
		var zoom = this.documentEditor.zoom();
		var cropButton = $(this.cropButton);
		cropButton.css("left", (this.activePageData.cropPos.x + newLeft) * zoom + "px");
		cropButton.css("top", (this.activePageData.cropPos.y + newTop) * zoom + "px");
	}

	if (newRect.width > 0 && newRect.height > 0) {
		this.updateRect(page, newRect, true);
	}

	return false;
};

batchApp.datacontext.CropEditor.prototype.createRect = function (point1, point2) {
	var left = Math.min(point1.x, point2.x);
	var right = Math.max(point1.x, point2.x);
	var top = Math.min(point1.y, point2.y);
	var bottom = Math.max(point1.y, point2.y);

	return {
		left: left,
		top: top,
		width: right - left,
		height: bottom - top
	};
};

batchApp.datacontext.CropEditor.prototype.onMouseUp = function (pageData) {
	if (!this.isEnabled || !this.activePageData) {
		return true;
	}

	this.documentEditor.releaseCapture();

	this.activePageData = null;
	return false;
};

/**
 * Обработать событие изменения страниц
 * @param {} pages Индексы повёрнутых страниц
 * @param {} rectTransformer Обработчик прямоугольника на странице
 * @param {} setChanged Отнести ли страницу к изменённым (или, наоборот, неизменённым)
 */
batchApp.datacontext.CropEditor.prototype.processPages = function (pages, rectTransformer, setChanged) {
	var self = this;

	if (!self.isEnabled) {
		return true;
	}

	var updates = [];

	// Необходимо аналогичным образом повернуть все прямоугольники
	this.rectangles.forEach(function (rect, pageId) {
		if (pages.indexOf(pageId) > -1) {
			var page = self.documentEditor.getPageById(pageId);
			if (page) {
				var transformedRect = rectTransformer(rect, page);
				if (setChanged) {
					self.changedIds.add(pageId);
				} else {
					self.changedIds.delete(pageId);
				}

				updates.push({ page: page, rect: transformedRect });
			}
		}
	});

	updates.forEach(function (data) {
		self.updateRect(data.page, data.rect, false);
	});

	return false;
};

/**
 * Обработать событие поворота страниц
 * @param {} angle Угол поворота
 * @param {} pages Индексы повёрнутых страниц
 */
batchApp.datacontext.CropEditor.prototype.onRotate = function (angle, pages) {
	return this.processPages(pages, function (rect, page) {
		return batchApp.datacontext.RectHelper.rotateRect(rect, angle, page);
	}, true);
};

/**
 * Обработать событие возврата страниц к исходному изображению
 * @param {} pages Индексы затронутых страниц
 */
batchApp.datacontext.CropEditor.prototype.onRestore = function (pages) {
	return this.processPages(pages, function (rect, page) {
		return {
			left: 0,
			top: 0,
			width: page.imageWidth,
			height: page.imageHeight
		};
	}, false);
};

/**
 * Обработать событие отражения
 * @param {} flipMode режим отражения
 * @param {} pages Индексы затронутых страниц
 */
batchApp.datacontext.CropEditor.prototype.onFlip = function (flipMode, pages) {
	return this.processPages(pages, function (rect, page) {
		return batchApp.datacontext.RectHelper.flipRect(rect, flipMode, page);
	}, true);
};

/**
 * Активировать или деактивировтаь редактор
 * @returns {} 
 */
batchApp.datacontext.CropEditor.prototype.toggle = function () {
	this.isEnabled = !this.isEnabled;

	if (this.isEnabled) {
		var pages = this.documentEditor.pages();
		var pageCount = pages.length;
		for (var i = 0; i < pageCount; i++) {
			var rect = {
				left: 0,
				top: 0,
				width: pages[i].imageWidth,
				height: pages[i].imageHeight
			};

			this.updateRect(pages[i], rect);
		}
	} else {
		var self = this;
		this.rectangles.forEach(function (rect, pageId) {
			var drawer = self.documentEditor.getPageDrawer(pageId, 1);
			drawer.clearDrawCommands();
			drawer.commit();
		});

		this.rectangles.clear();
		this.changedIds.clear();

		self.documentEditor.setCursor("auto");

		this.cropButton.style.display = 'none';
		this.onDeactivate();
	}

	return this.isEnabled;
};

batchApp.datacontext.CropEditor.prototype.updateRect = function (page, rect, redraw) {
	var pageId = page.id;
	this.rectangles.set(pageId, rect);

	var drawer = this.documentEditor.getPageDrawer(pageId, 1);
	drawer.clearDrawCommands();

	// Нарисуем фон
	drawer.addDrawCommand(FC.DefaultDrawer.DrawCommands.FillStyle, 'rgba(2, 46, 79, 0.4)');
	drawer.addDrawCommand(FC.DefaultDrawer.DrawCommands.BeginPath);
	drawer.addDrawCommand(FC.DefaultDrawer.DrawCommands.Rect, rect.left, rect.top, rect.width, rect.height);
	drawer.addDrawCommand(FC.DefaultDrawer.DrawCommands.Rect, page.imageWidth, 0, -page.imageWidth, page.imageHeight);
	drawer.addDrawCommand(FC.DefaultDrawer.DrawCommands.Fill);

	var self = this;
	drawer.addCustomDrawCommand(function (context) {
		var dash;

		var lineLength = batchApp.datacontext.CropEditor.LineLength / self.documentEditor.zoom();
		var horizontalSpace = rect.width / 2 - lineLength * 1.5;
		var verticalSpace = rect.height / 2 - lineLength * 1.5;

		// Штриховку нужно подбирать для разных вариантов прямоугольника
		if (horizontalSpace > 0 && verticalSpace > 0) {
			dash = [
				lineLength, horizontalSpace, lineLength, horizontalSpace, lineLength * 2,
				verticalSpace, lineLength, verticalSpace, lineLength * 2,
				horizontalSpace, lineLength, horizontalSpace, lineLength * 2,
				verticalSpace, lineLength, verticalSpace, lineLength
			];
		} else if (horizontalSpace > 0) {
			dash = [lineLength, horizontalSpace, lineLength, horizontalSpace, lineLength * 2 + rect.height,
				horizontalSpace, lineLength, horizontalSpace, lineLength + rect.height];
		} else if (verticalSpace > 0) {
			dash = [rect.width + lineLength, verticalSpace, lineLength, verticalSpace, lineLength * 2 + rect.width, verticalSpace, lineLength, verticalSpace, lineLength];
		} else {
			dash = [1, 0];
		}

		context.strokeStyle = 'black';
		context.lineWidth = batchApp.datacontext.CropEditor.LineThickness / self.documentEditor.zoom();
		context.setLineDash(dash);

		context.strokeRect(rect.left, rect.top, rect.width, rect.height);
	});

	drawer.commit();

	if (redraw) {
		this.changedIds.add(pageId);
	}
};