﻿/**
 * Редактор Redact для DocumentEditor (расширение)
 * @returns {} 
 */
batchApp.datacontext.RedactEditor = function (documentEditor, lock, dataContext, onDeactivate, onUpdatePage) {
	var self = this;

	self.documentEditor = documentEditor;
	self.lock = lock;
	self.dataContext = dataContext;
	self.onDeactivate = onDeactivate;
	self.onUpdatePage = onUpdatePage;
	self.isEnabled = false;
	self.rectangles = new Map();
	self.activePageData = null;

	self.documentEditor.addOnMouseDown(self.onMouseDown.bind(this));
	self.documentEditor.addOnMouseMove(self.onMouseMove.bind(this));
	self.documentEditor.addOnMouseUp(self.onMouseUp.bind(this));

	self.documentEditor.onOpenedDocumentChanged(function () {
	    if (self.isEnabled) {
	        self.cancel();
	    }
	});
};

/**
 * Отменить Redact
 * @returns {} 
 */
batchApp.datacontext.RedactEditor.prototype.cancel = function () {
	var self = this;
	self.isEnabled = false;

	self.rectangles.forEach(function(rectSet, pageId) {
		var drawer = self.documentEditor.getPageDrawer(pageId, 1);
		drawer.clearDrawCommands();
		drawer.commit();
	});

	self.onDeactivate();

	self.rectangles.clear();
};

/**
 * Зафиксировать Redact
 * @param {any} callback Callback вызываемый при полном завершении redact
 */
batchApp.datacontext.RedactEditor.prototype.redact = function (callback) {
	var self = this;

	if (self.lock()) {
		return;
	}

	var total = 0;
	this.rectangles.forEach(function (rectSet, pageId) {
		total += rectSet.size;
	});

	var count = 0;

	if (this.rectangles.size === 0) {
		if (callback) {
			callback()
        }
	}
	this.rectangles.forEach(function (rectSet, pageId) {
		var page = self.documentEditor.getPageById(pageId);
		var rectsCount = rectSet.size;
		var localCounter = 0;

		rectSet.forEach(function(rect) {
			self.dataContext.redactPage(pageId, Math.round(rect.left), Math.round(rect.top),
				Math.round(rect.left + rect.width), Math.round(rect.top + rect.height), function (result) {
					count++;
					if (count === total) {
						self.rectangles.clear(); // Redact закончен полностью
					}

					if (result.hasErrors) {
						FC.Utils.Error.Show(result.errorText);
						return;
					}

					localCounter++;
					if (localCounter === rectsCount) { // Обновим страницу только тогда, когда будут применены все прямоугольники
						self.onUpdatePage(pageId);
						// Вызываем callback при полном обновлении элементов redact
						if (callback) {
							callback();
						}
					}
				});
		});
	});
};

batchApp.datacontext.RedactEditor.prototype.onMouseDown = function (pageData) {
	if (!this.isEnabled) {
		return true;
	}

	var page = this.documentEditor.getPageById(pageData.pageId);
	if (!page) {
		return true;
	}

	if (pageData.x < 0 || pageData.x > page.imageWidth || pageData.y < 0 || pageData.y > page.imageHeight) {
		return true;
	}

	var rectSet;
	if (!this.rectangles.has(pageData.pageId)) {
		rectSet = new Set();
		this.rectangles.set(pageData.pageId, rectSet);
	} else {
		rectSet = this.rectangles.get(pageData.pageId);

		// Проверим, не нажали ли мы на крестик
		var activeRect = this.getActiveRect(pageData);
		if (activeRect) {
			rectSet.delete(activeRect);
			this.updatePage(page);
			if (rectSet.size === 0) {
				this.rectangles.delete(pageData.pageId)
			}
			return false;
		}
	}

	var newRect = {
		left: pageData.x,
		top: pageData.y,
		width: 0,
		height: 0,
		isFinished: false
	};

	rectSet.add(newRect);

	this.updatePage(page);
	this.activePageData = {
		pageId: pageData.pageId,
		x: pageData.x,
		y: pageData.y,
		rect: newRect
	};

	this.documentEditor.captureMouse(pageData.pageId);
	return false;
};

/**
 * Обработать событие изменения страниц
 * @param {} pages Индексы повёрнутых страниц
 * @param {} rectTransformer Обработчик прямоугольника на странице
 */
batchApp.datacontext.RedactEditor.prototype.processPages = function (pages, rectTransformer) {
	var self = this;

	if (!self.isEnabled) {
		return true;
	}

	var updates = [];

	// Необходимо аналогичным образом повернуть все прямоугольники
	this.rectangles.forEach(function (rectSet, pageId) {
		if (pages.indexOf(pageId) > -1) {
			var page = self.documentEditor.getPageById(pageId);
			if (page) {
				var newRectSet = new Set();

				rectSet.forEach(function (rect) {
					var transformedRect = rectTransformer(rect, page);
					transformedRect.isFinished = rect.isFinished;
					newRectSet.add(transformedRect);
				});

				updates.push({ page: page, rectSet: newRectSet });
			}
		}
	});

	updates.forEach(function (item) {
		self.rectangles.set(item.page.pageData.Id(), item.rectSet);
		self.updatePage(item.page);
	});

	self.documentEditor.Draw();

	return false;
};

/**
 * Обработать событие поворота страниц
 * @param {} angle Угол поворота
 * @param {} pages Индексы повёрнутых страниц
 */
batchApp.datacontext.RedactEditor.prototype.onRotate = function (angle, pages) {
	return this.processPages(pages, function (rect, page) {
		return batchApp.datacontext.RectHelper.rotateRect(rect, angle, page);
	});
};

/**
 * Обработать событие возврата страниц к исходному изображению
 * @param {} pages Индексы затронутых страниц
 */
batchApp.datacontext.RedactEditor.prototype.onRestore = function (pages) {
	var self = this;

	if (!self.isEnabled) {
		return true;
	}

	// Удаляем прямоугольники с затронутых страниц
	pages.forEach(function (pageId) {
		var page = self.documentEditor.getPageById(pageId);
		if (page) {
			self.rectangles.delete(pageId);
			self.updatePage(page);
		}
	});

	return true;
};

/**
 * Обработать событие отражения
 * @param {} flipMode режим отражения
 * @param {} pages Индексы затронутых страниц
 */
batchApp.datacontext.RedactEditor.prototype.onFlip = function (flipMode, pages) {
	return this.processPages(pages, function (rect, page) {
		return batchApp.datacontext.RectHelper.flipRect(rect, flipMode, page);
	});
};

/**
 * Проверить попадание точки внутрь прямоугольника
 * @param {} rect 
 * @param {} point 
 * @returns {} 
 */
batchApp.datacontext.RedactEditor.prototype.isInsideRect = function (rect, point) {
	var right = rect.left + rect.width;
	var bottom = rect.top + rect.height;

	return rect.left <= point.x &&
		rect.top <= point.y &&
		right >= point.x &&
		bottom >= point.y;
}

/**
 * Проверить, на находится ли мышь в зоне крестика. И если находится, вернуть соответствующий прямоугольник
 * @param {} pageData 
 * @returns {} 
 */
batchApp.datacontext.RedactEditor.prototype.getActiveRect = function (pageData) {

	var rectSet = this.rectangles.get(pageData.pageId);
	var zoom = this.documentEditor.zoom();

	var margin = batchApp.datacontext.RedactEditor.CrossMargin / zoom;
	var self = this;
	var activeRect;
	rectSet.forEach(function (rect) {
		if (!rect.isFinished) {
			return;
		}

		var cornerLeft = rect.left + rect.width - batchApp.datacontext.RedactEditor.CrossSize / zoom - margin * 2;

		var crossRect = {
			left: cornerLeft,
			top: rect.top,
			width: batchApp.datacontext.RedactEditor.CrossSize / zoom + margin * 2,
			height: batchApp.datacontext.RedactEditor.CrossSize / zoom + margin * 2
		};

		if (self.isInsideRect(crossRect, pageData)) {
			activeRect = rect;
		}
	});

	return activeRect;
}

batchApp.datacontext.RedactEditor.prototype.onMouseMove = function(pageData) {
	if (!this.isEnabled) {
		return true;
	}

	var page = this.documentEditor.getPageById(pageData.pageId);
	if (!page) {
		return true;
	}

	if (!this.activePageData) {
		// Изменим курсор при движении над крестиком
		if (this.rectangles.has(pageData.pageId)) {

			var activeRect = this.getActiveRect(pageData);

			if (activeRect) {
				this.documentEditor.setCursor('pointer');
			} else {
				this.documentEditor.setCursor('auto');
			}
		}

		return true;
	}

	if (this.activePageData.pageId !== pageData.pageId) {
		return true;
	}

	if (pageData.x < 0) {
		pageData.x = 0;
	} else if (pageData.x > page.imageWidth) {
		pageData.x = page.imageWidth;
	}

	if (pageData.y < 0) {
		pageData.y = 0;
	} else if (pageData.y > page.imageHeight) {
		pageData.y = page.imageHeight;
	}

	var currentRect = this.activePageData.rect;

	currentRect.left = Math.min(this.activePageData.x, pageData.x);
	currentRect.top = Math.min(this.activePageData.y, pageData.y);
	currentRect.width = Math.abs(this.activePageData.x - pageData.x);
	currentRect.height = Math.abs(this.activePageData.y - pageData.y);

	this.updatePage(page);
	return false;
};

batchApp.datacontext.RedactEditor.prototype.onMouseUp = function (pageData) {
	if (!this.isEnabled || !this.activePageData) {
		return true;
	}

	this.documentEditor.releaseCapture();

	var activeRect = this.activePageData.rect;

	if (activeRect.width > 0 && activeRect.height > 0) {
		activeRect.isFinished = true;

		var page = this.documentEditor.getPageById(this.activePageData.pageId);
		this.updatePage(page);
	} else {
		// Удалим прямоугольник
		var rectSet = this.rectangles.get(pageData.pageId);
		rectSet.delete(activeRect);
	}

	this.activePageData = null;
	return false;
};

/**
 * Активировать или деактивировтаь редактор
 * @returns {} 
 */
batchApp.datacontext.RedactEditor.prototype.toggle = function () {
	this.isEnabled = !this.isEnabled;

	if (!this.isEnabled) {
		this.complete();
	}

	return this.isEnabled;
};

/**
 * Закомитить и зафиксировать редактор
 * @returns {} 
 */
batchApp.datacontext.RedactEditor.prototype.complete = function (callback) {
	var pages = this.documentEditor.pages();
	var pageCount = pages.length;
	for (var i = 0; i < pageCount; i++) {
			var pageId = pages[i].id;

			var drawer = this.documentEditor.getPageDrawer(pageId, 1);
			drawer.clearDrawCommands();
			drawer.commit();
		}
	this.redact(callback);
};

batchApp.datacontext.RedactEditor.CrossSize = 7;
batchApp.datacontext.RedactEditor.CrossThickness = 1;
batchApp.datacontext.RedactEditor.CrossMargin = 3;

batchApp.datacontext.RedactEditor.prototype.updatePage = function (page) {
	var pageId = page.id;

	var rectSet = this.rectangles.get(pageId);

	var drawer = this.documentEditor.getPageDrawer(pageId, 1);
	drawer.clearDrawCommands();

	if (!rectSet) {
		return;
	}

	rectSet.forEach(function (rect) {
		drawer.addDrawRectangleCommand({
			rectangle: rect,
			fill: {
				color: 'black'
			}
		});
	});

	var self = this;
	rectSet.forEach(function (rect) {
		// Крестик в правом верхнем углу
		if (rect.isFinished) {
			drawer.addCustomDrawCommand(function (context) {

				var margin = batchApp.datacontext.RedactEditor.CrossMargin / self.documentEditor.zoom();
				var crossSize = Math.min(rect.width - margin,
					Math.min(rect.height - margin,
						batchApp.datacontext.RedactEditor.CrossSize / self.documentEditor.zoom()));
				var thickness = batchApp.datacontext.RedactEditor.CrossThickness / self.documentEditor.zoom();
				var cornerLeft = rect.left + rect.width - crossSize - margin;

				context.strokeStyle = 'white';
				context.lineWidth = thickness;
				context.beginPath();
				context.moveTo(cornerLeft, rect.top + margin);
				context.lineTo(rect.left + rect.width - margin, rect.top + crossSize + margin);
				context.moveTo(cornerLeft, rect.top + crossSize + margin);
				context.lineTo(rect.left + rect.width - margin, rect.top + margin);
				context.closePath();
				context.stroke();
			});
		}
	});

	drawer.commit();
};