﻿FC = FC || {};

/**
 * Класс для автоматизированного манипулирования пакетами
 * @returns {} 
 */
FC.BatchesManager = function (batchesDataContext, batchDataContext, env, onOpenBatch, onError) {
	this.batchesDataContext = batchesDataContext;
	this.batchDataContext = batchDataContext;
	this.onOpenBatch = onOpenBatch;
	this.env = env;
	this.onError = onError;
};

/**
 * Создать или открыть пакет
 * @param  settings Настройки создания пакета
 * @returns {} 
 */
FC.BatchesManager.prototype.openOrCreate = function (settings) {
	var self = this;
	self.findBatch(settings, function (batchId) {
		if (batchId > -1) {
			self.open(settings, batchId);
		} else {
			self.create(settings);
		}
	});
};

/**
 * Найти и открыть подходящий пакет
 * @param  settings Настройки создания пакета
 * @returns {} 
 */
FC.BatchesManager.prototype.findAndOpen = function (settings) {
	var self = this;
	self.findBatch(settings, function (batchId) {
		if (batchId > -1) {
			self.open(settings, batchId);
		} else {
			self.onError(self.env.Localization.GlobalResources.CannotFindBatch);
		}
	});
};

/**
 * Создать пакет
 * @param {} settings Настройки создания пакета
 * @returns {} 
 */
FC.BatchesManager.prototype.create = function (settings) {
	var self = this;

	self.setBatchType(settings,
		function () {
			self.batchesDataContext.createBatch(settings.BatchType,
				function (creationResult) {
					if (creationResult.hasErrors) {
						FC.Utils.Error.Show(creationResult.errorText);
					} else {
						var batchId = parseInt(creationResult.result);
						self.open(settings, batchId, true);
					}
				});
		});
};

/**
 * Получить пакет пересканирования
 * @param {} settings 
 * @returns {} 
 */
FC.BatchesManager.prototype.getRescan = function (settings) {
	var rescanTasks = ko.observableArray();
	var self = this;
	self.batchesDataContext.getRescanTasks(rescanTasks,
		function () {
			var ids = self.filterBatches(rescanTasks(), -1, true);
			var batchId = ids[0];
			var rescanId = ids[1];

			if (batchId > -1) {
				self.open(settings, batchId);
			} else if (rescanId > -1) {
				self.batchesDataContext.openRescanBatch(rescanId, function (openRescanResult) {
					if (openRescanResult.hasErrors) {
						self.onError(openRescanResult.errorText);
					} else {
						self.open(settings, openRescanResult.result);
					}
				});
			} else {
				self.onError(self.env.Localization.GlobalResources.CannotFindBatch);
			}
		});
};

/**
 * Функция проверки статуса пакета
 * @param {} status Статус пакета
 * @param {} rescan Является ли пакет пакетом пересканирования
 * @returns {} Подходит ли пакет
 */
FC.BatchesManager.prototype.statusCheck = function(status, rescan) {
	return status === 0 || status === 1 || status === 4 && rescan;
}

/**
 * Выбрать подходящий пакет из списка
 * @param {} batches Набор пакетов 
 * @param {} batchTypeId Идентификатор типа пакета
 * @param {} rescan Является ли пакет пакетом пересканирования
 * @returns {} Идентификатор найденного пакета или -1
 */
FC.BatchesManager.prototype.filterBatches = function (batches, batchTypeId, rescan) {
	var self = this;
	var batchId = -1;
	var rescanId = -1;

	$.each(batches,
		function (key, value) {
			if (batchTypeId > -1 && parseInt(value.BatchTypeId()) !== batchTypeId) {
				return true;
			}

			var status = value.Status();
			if (self.statusCheck(status, rescan)) {
				batchId = value.Id();
				rescanId = value.RescanTaskID();
				return false;
			}

			return true;
		});

	return [batchId, rescanId];
}

/**
 * Найти пакет, подходящий под условия (batchTypeId и статус)
 * @param {} settings 
 * @param callbackFunc Функция, выполняемая далее
 * @returns {} 
 */
FC.BatchesManager.prototype.findBatch = function (settings, callbackFunc) {
	var self = this;
	self.setBatchType(settings,
		function () {
			var availableBatches = ko.observableArray();

			self.batchesDataContext.getBatches(availableBatches,
				function () {
					var batchId = self.filterBatches(availableBatches(), settings.BatchType ? settings.BatchType.Id() : -1)[0];
					callbackFunc(batchId);
				});
		});
};

/**
 * Открыть пакет
 * @param {} settings Настройки создания пакета
 * @param {} batchId Идентификатор пакета
 * @param {} isNew Является ли этот пакет только что созданным (можно задать readonly-регпараметры)
 * @returns {} 
 */
FC.BatchesManager.prototype.open = function (settings, batchId, isNew) {

	var self = this;
	var openFunc = function () {
		self.onOpenBatch(batchId);
	};

	var setRegParamsFunc = function() {
		if (Object.keys(settings.RegParams).length > 0) {
			self.setRegParams(batchId, settings.RegParams, isNew, openFunc);
		} else {
			openFunc();
		}
	};

	if (settings.BatchName) {
		self.renameBatch(batchId, settings.BatchName, setRegParamsFunc);
	} else {
		setRegParamsFunc();
	}
};

/**
 * Инициализировать пакет
 * @param {} batchId Идентификатор пакета
 * @param {} callbackFunc Функция, выполняемая далее
 * @returns {} 
 */
FC.BatchesManager.prototype.initBatch = function(batchId, callbackFunc) {
	var self = this;
	if (self.batchDataContext.inited()) {
		callbackFunc();
		return;
	}

	self.batchDataContext.inited.subscribe(callbackFunc);
	self.batchDataContext.initModule(batchId);
};

/**
 * Переименовать пакет
 * @param {} batchId Идентификатор пакета
 * @param {} batchName Новое имя пакета
 * @param {} callbackFunc Функция, выполняемая далее
 * @returns {} 
 */
FC.BatchesManager.prototype.renameBatch = function (batchId, batchName, callbackFunc) {
	var self = this;
	self.initBatch(batchId,
		function () {
			self.batchDataContext.setBatchName(batchName, callbackFunc);
		});
};

/**
 * Установить регистрационные параметры для текущего пакета
 * @param batchId Идентификатор пакета
 * @param regParams Задаваемые регистрационные параметры
 * @param {} isNew Является ли этот пакет только что созданным (можно задать readonly-регпараметры)
 * @param callbackFunc Функция, выполняемая далее
 * @returns {} 
 */
FC.BatchesManager.prototype.setRegParams = function (batchId, regParams, isNew, callbackFunc) {
	var self = this;
	self.initBatch(batchId, function () {
		var batchRegParams = ko.observableArray([]);
		self.batchDataContext.getBatchRegParameters(batchRegParams,
			function () {
				var changed = false;

				for (var paramName in regParams) {
					if (!regParams.hasOwnProperty(paramName)) {
						continue;;
					}

					var paramValue = regParams[paramName];

					$.each(batchRegParams(),
						function (keyParam, param) {
							var name = param.Name();
							if (name === paramName) {
								if (isNew || !param.IsReadOnly()) {
									param.Value(paramValue);
									changed = true;
								}

								return false;
							}

							return true;
						});
				}

				if (changed) {
					self.batchDataContext.saveBatchRegParameters(batchRegParams(), callbackFunc);
				} else {
					callbackFunc();
				}
			});
	});
}

/**
 * Получить тип пакета по имени
 * @param callbackFunc Функция, выполняемая далее
 */
FC.BatchesManager.prototype.getBatchTypeByName = function (batchTypeName, callbackFunc) {
	if (!batchTypeName) {
		callbackFunc(null);
		return;
	}

	this.batchesDataContext.getBatchTypes(function(batchesType) {
		if (batchesType == null)
			return;

		var foundBatchType = null;
		$.each(batchesType,
			function(key, batchType) {
				if (batchType.Name() === batchTypeName || (batchType.Id() === 0 && batchTypeName.toLowerCase() === "default")) {
					foundBatchType = batchType;
					return false;
				}

				return true;
			});

		if (callbackFunc) {
			callbackFunc(foundBatchType);
		}
	});
};

/**
 * Задать информацию о типе пакета, если она ещё не задана
 * @param {} settings Настройки
 * @param {} callbackFunc Функция, выполняемая далее
 * @returns {} 
 */
FC.BatchesManager.prototype.setBatchType = function(settings, callbackFunc) {
	if (settings.BatchType || !settings.BatchTypeName) {
		callbackFunc();
	} else {
		var self = this;
		this.getBatchTypeByName(settings.BatchTypeName,
			function (batchType) {
				if (!batchType) {
					self.onError(self.env.Localization.GlobalResources.CannotFindBatchType + ' "' + settings.BatchTypeName + '"!');
					return;
				}

				settings.BatchType = batchType;
				callbackFunc();
			});
	}
};