﻿/**
 * Базовая зона управления размером полигона
 * @param {} point точка, от которой рисуется зона
 * @param {} rectPosition положение квадратика управления
 * @param {} influencedValues значения, на которые влияет перемещение зоны
 * @param {} canChangeHorizontalValue можно ли менять горизонтальное значение
 * @param {} canChangeVerticalValue можно ли менять вертикальное значение
 * @param {} target регион, с которым связана зона
 * @param {} page страница, на которой существует регион
 * @param {} depth грубина зоны
 * @param {} cursor курсор
 */
FC.Viewmodels.ActionZones.PolygonBorderZone = function (point, rectPosition, influencedValues, canChangeHorizontalValue, canChangeVerticalValue, target, page, depth, cursor) {

	// вычислим, какая должна быть активная зона у этой зоны
	var rectangles = ko.pureComputed(function () {
		var indentSmall = (FC.Viewmodels.ActionZones.zoneConstants.borderZonesPadInside) / page.zoom();
		var indentBig = (FC.Viewmodels.ActionZones.zoneConstants.borderZoneSize -
			FC.Viewmodels.ActionZones.zoneConstants.borderZonesPadInside) / page.zoom();
		var indentHalf = (FC.Viewmodels.ActionZones.zoneConstants.borderZoneSize * 0.5) / page.zoom();

		var rectPos = ko.utils.unwrapObservable(rectPosition);
		var positionH = rectPos % 10;
		var positionV = (Math.floor(rectPos / 10)) % 10;

		return [new FC.Viewmodels.ImageEditor.Rectangle(
			{
				x1: point.x() - (positionH === 0 ? indentHalf : (positionH === 1 ? indentSmall : indentBig)),
				y1: point.y() - (positionV === 0 ? indentHalf : (positionV === 1 ? indentSmall : indentBig)),
				x2: point.x() + (positionH === 0 ? indentHalf : (positionH === 1 ? indentBig : indentSmall)),
				y2: point.y() + (positionV === 0 ? indentHalf : (positionV === 1 ? indentBig : indentSmall))
			})];
	});

	// наследуем от базовой зоны
	FC.Viewmodels.ActionZones.BaseZone.call(this, rectangles, target, page, FC.Viewmodels.ActionZones.zoneActivationType.selected);

	// переопределяем курсор.
	if (!ko.isComputed(cursor)) {
		this.cursorStyle(cursor);
	} else {
		this.cursorStyle = cursor;
	}

	this.point = point;
	this.influencedValues = influencedValues;
	this.depth(depth + 61);
	this.startPoint = null;
	this.canChangeHorizontalValue = canChangeHorizontalValue;
	this.canChangeVerticalValue = canChangeVerticalValue;
};

FC.Viewmodels.ActionZones.PolygonBorderZone.prototype = Object.create(FC.Viewmodels.ActionZones.BaseZone.prototype);

/**
 * Нарисовать зону
 * @param {} layerIndex индекс текущего слоя рисования
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.prototype.draw = function (layerIndex) {
	// рисуем только активные зоны и только на втором слое
	if (!this.isActive() || layerIndex === 0 || !this.target.field()) {
		return;
	}

	// у разных типов регионов - разное отображение квадратиков
	var isImage = this.target.field().fieldType === FC.BusinessModels.DocumentEditor.FieldTypes.Picture;
	var isCheckboxGroup = this.target.field().fieldType === FC.BusinessModels.DocumentEditor.FieldTypes.Choice;
	var isGroup = this.target.field().fieldType === FC.BusinessModels.DocumentEditor.FieldTypes.GroupParent;

	var cssClass = ".region-squares";

	if (isImage) {
		cssClass = ".regionImage-squares";
	} else if (isCheckboxGroup) {
		cssClass = ".regionChoice-squares";
	} else if (isGroup) {
		cssClass = ".regionGroup-squares";
	}

	var squaresColor = FC.Utils.GetCssClassValue(cssClass, "background-color");
	var layer = this.page.getLayer(1);
	var rects = ko.utils.unwrapObservable(this.rectangles);
	layer.drawFilledRectangle(rects[0], squaresColor);
};

/**
 * ключ, по которому можно удалить события зоны из DocumentEditor
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey = "zonesResizing";

/**
 * Нажатие мыши
 * @param {} point точка
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.prototype.mouseDown = function (point) {
	var self = this;

	if (!self.isActive() || !self.pointIsInside(point)) {
		return false;
	}

	self.startPoint = point;

	var mouseMove = function (data) { self.mouseMove(self, data); };
	var mouseUp = function () { self.mouseUp(self); };

	// привязываемся к событиям DocumentEditor
	var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();
	docEditor.addOnMouseUp(mouseUp.bind(this), 100, FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey);
	docEditor.addOnMouseLeave(mouseUp.bind(this), 100, FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey);
	docEditor.addOnMouseMove(mouseMove.bind(this), 100, FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey);

	return true;
};

/**
 * Отжатие мыши
 * @param {} zone зона, на которой мышь нажали
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.prototype.mouseUp = function (zone) {
	zone.startPoint = null;

	// отвязываемся от событий DocumentEditor
	var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();
	docEditor.removeOnMouseUp(FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey);
	docEditor.removeOnMouseLeave(FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey);
	docEditor.removeOnMouseMove(FC.Viewmodels.ActionZones.PolygonBorderZone.handlerRemovalKey);

	if (zone.target.resized) {
		zone.target.updateText();
	}
};

/**
 * Перемещение мыши
 * @param {} zone  зона, на которой мышь нажали
 * @param {} point точка 
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.prototype.mouseMove = function (zone, point) {

	// считаем, на сколько сдвинулась мышь
	var deltaX = Math.round(point.x - zone.startPoint.x);
	var deltaY = Math.round(point.y - zone.startPoint.y);
	var influenced = ko.utils.unwrapObservable(this.influencedValues);

	// определяем, можем ли мы изменять горизонтальное и вертикальное значение
	var canChangeX = this.canChangeHorizontalValue(deltaX);
	var canChangeY = this.canChangeVerticalValue(deltaY);

	if (canChangeX || canChangeY) {
		// изменяем все значения, на которые влияет текущая зона
		for (var i = 0; i < influenced.length; i++) {
			if (influenced[i].isHorizontal && canChangeX) {
				influenced[i].addToValue(deltaX);
			}

			if (!influenced[i].isHorizontal && canChangeY) {
				influenced[i].addToValue(deltaY);
			}
		}

		zone.target.updateText();
		FC.Viewmodels.EventStack.addEvent();

		if (canChangeX) {
			zone.startPoint.x = point.x;
		}

		if (canChangeY) {
			zone.startPoint.y = point.y;
		}

		zone.page.draw();
	}
};

/**
 * Можно ли менять горизонтальное значение
 * @param {} currentPoint текущая точка
 * @param {} allPoints все точки полигона
 * @param {} currentIndex индекс текущей точки
 * @param {} delta дельта, на которую хотим менять
 * @param {} page текущая страница
 * @returns {boolean} 
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.canChangeHorizontalValue = function (currentPoint, allPoints, currentIndex, delta,
	page) {
	if (delta === 0) {
		return false;
	}

	// находим точку до и после текущей
	// а также точку после следующей
	var pointBefore = allPoints[currentIndex > 0 ? currentIndex - 1 : allPoints.length - 1];
	var pointNext = allPoints[currentIndex < allPoints.length - 1 ? currentIndex + 1 : 0];
	var pointAfterNext = allPoints[currentIndex < allPoints.length - 2 ? currentIndex + 2 : currentIndex - allPoints.length + 2];

	var xBefore = pointBefore.x();
	var xAfter = pointAfterNext.x();

	var x = currentPoint.x();

	var currentY = currentPoint.y();
	var nextY = xBefore === x ? pointBefore.y() : pointNext.y();
	var currentY1 = Math.min(currentY, nextY);
	var currentY2 = Math.max(currentY, nextY);

	var i;

	var previousPointY;
	var y;
	var nextPointY;
	var otherY;
	var y1;
	var y2;

	// если уменьшаем значение
	if (delta < 0) {
		var xMax = 0;

		// если мы двигаем правые точки, то должны упереться в самую правую из левых
		// на интервале y1,y2
		if (currentPoint.pointType === 1) {
			for (i = 0; i < allPoints.length; i++) {
				if (allPoints[i].pointType === 1) {
					continue;
				}

				y = allPoints[i].y();
				previousPointY = allPoints[i > 0 ? i - 1 : allPoints.length - 1].y();
				nextPointY = allPoints[i < allPoints.length - 1 ? i + 1 : 0].y();
				otherY = previousPointY === y ? nextPointY : previousPointY;
				y1 = Math.min(y, otherY);
				y2 = Math.max(y, otherY);

				if (allPoints[i].x() > xMax && currentY1 <= y2 && currentY2 >= y1) {
					xMax = allPoints[i].x();
				}
			}
		}

		// если x - самая левая из трех точек
		if (xBefore >= x && xAfter >= x) {
			// если мы слева - то ограничены началом страницы
			if (currentPoint.pointType === 0) {
				return x + delta > 0;
			}

			// если справа - то xMax
			return currentPoint.x() + delta > xMax;
		}

		if (xBefore < x && xAfter < x) {
			return x + delta > Math.max(xBefore, xAfter, xMax);
		}

		if (xBefore < x) {
			return x + delta > Math.max(xBefore, xMax);
		}

		return x + delta > Math.max(xAfter, xMax);
	} else {
		// если увеличиваем значение
		var xMin = page.width();

		// если текущая точка - слева,
		// то при увеличении значения мы должны упереться в самую левую
		// точку из правых, находящихся напротив текущей
		if (currentPoint.pointType === 0) {
			for (i = 0; i < allPoints.length; i++) {
				if (allPoints[i].pointType === 0) {
					continue;
				}

				y = allPoints[i].y();
				previousPointY = allPoints[i > 0 ? i - 1 : allPoints.length - 1].y();
				nextPointY = allPoints[i < allPoints.length - 1 ? i + 1 : 0].y();
				otherY = previousPointY === y ? nextPointY : previousPointY;
				y1 = Math.min(y, otherY);
				y2 = Math.max(y, otherY);

				if (allPoints[i].x() < xMin && currentY1 <= y2 && currentY2 >= y1) {
					xMin = allPoints[i].x();
				}
			}
		}


		if (xBefore <= x && xAfter <= x) {
			if (currentPoint.pointType === 1) {
				return x + delta < page.width();
			}

			return x + delta < xMin;
		}

		if (xBefore > x && xAfter > x) {
			return x + delta < Math.min(xBefore, xAfter, xMin);
		}

		if (xBefore > x) {
			return x + delta < Math.min(xBefore, xMin);
		}


		return x + delta < Math.min(xAfter, xMin);
	}
};

/**
 * Можно ли менять вертикальное значение
 * @param {} currentPoint текущая точка
 * @param {} allPoints все точки полигона
 * @param {} currentIndex индекс текущей точки
 * @param {} delta дельта, на которую хотим менять
 * @param {} page текущая страница
 * @returns {boolean} 
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.canChangeVerticalValue = function (currentPoint, allPoints, currentIndex, delta,
	page) {
	if (delta === 0) {
		return false;
	}

	// находим точку до и после текущей
	// а также точку после следующей
	var pointBefore = allPoints[currentIndex > 0 ? currentIndex - 1 : allPoints.length - 1];
	var pointNext = allPoints[currentIndex < allPoints.length - 1 ? currentIndex + 1 : 0];
	var pointAfter = allPoints[currentIndex < allPoints.length - 2 ? currentIndex + 2 : currentIndex - allPoints.length + 2];

	var yBefore = pointBefore.y();
	var yAfter = pointAfter.y();
	var y = currentPoint.y();

	var currentX = currentPoint.x();
	var nextX = yBefore === y ? pointBefore.x() : pointNext.x();
	var currentX1 = Math.min(currentX, nextX);
	var currentX2 = Math.max(currentX, nextX);

	var i;

	var previousPointX;
	var x;
	var nextPointX;
	var otherX;
	var x1;
	var x2;

	if (delta < 0) {
		// если уменьшаем значение
		var yMax = 0;

		// находим y, который меньше текущего и находится на том же уровне по x
		for (i = 0; i < allPoints.length; i++) {
			if (allPoints[i].y() >= y) {
				continue;
			}

			x = allPoints[i].x();
			previousPointX = allPoints[i > 0 ? i - 1 : allPoints.length - 1].x();
			nextPointX = allPoints[i < allPoints.length - 1 ? i + 1 : 0].x();
			otherX = previousPointX === x ? nextPointX : previousPointX;
			x1 = Math.min(x, otherX);
			x2 = Math.max(x, otherX);

			if (allPoints[i].y() > yMax && currentX1 <= x2 && currentX2 >= x1) {
				yMax = allPoints[i].y();
			}
		}


		if (yBefore >= y && yAfter >= y) {
			return y + delta > yMax;
		}

		if (yBefore < y && yAfter < y) {
			return y + delta > Math.max(yBefore, yAfter, yMax);
		}

		if (yBefore < y) {
			return y + delta > Math.max(yBefore, yMax);
		}

		return y + delta > Math.max(yAfter, yMax);
	} else {
		// если увеличиваем значение
		var yMin = page.height();

		// находим точку, у которой y>текущего и
		// которая находится на том же уровне по горизонтали
		for (i = 0; i < allPoints.length; i++) {
			if (allPoints[i].y() <= y) {
				continue;
			}

			x = allPoints[i].x();
			previousPointX = allPoints[i > 0 ? i - 1 : allPoints.length - 1].x();
			nextPointX = allPoints[i < allPoints.length - 1 ? i + 1 : 0].x();
			otherX = previousPointX === x ? nextPointX : previousPointX;
			x1 = Math.min(x, otherX);
			x2 = Math.max(x, otherX);

			if (allPoints[i].y() < yMin && currentX1 <= x2 && currentX2 >= x1) {
				yMin = allPoints[i].y();
			}
		}

		if (yBefore <= y && yAfter <= y) {
			return y + delta < Math.min(yMin, page.height());
		}

		if (yBefore > y && yAfter > y) {
			return y + delta < Math.min(yBefore, yAfter, yMin);
		}

		if (yBefore > y) {
			return y + delta < Math.min(yBefore, yMin);
		}

		return y + delta < Math.min(yAfter, yMin);
	}
};

/**
 * Возможное расположение квадратика управления по отношению к текущей точке.
 */
FC.Viewmodels.ActionZones.PolygonBorderZone.RectanglePositions = {
	SouthEast: 11,
	East: 01,
	NorthEast: 21,
	North: 20,
	NorthWest: 22,
	West: 02,
	SouthWest: 12,
	South: 10
};