﻿var FC = FC || {};
FC.Viewmodels = FC.Viewmodels || {};
FC.Viewmodels.DocumentEditor = FC.Viewmodels.DocumentEditor || {};

/**
 * Фабрика вьюмоделей регионов
 */
FC.Viewmodels.DocumentEditor.regionFactory = {

	/**
	 * Нужно ли создавать вьюмодель региона
	 * @param {} region модель региона
	 * @param {} page текущая страница
	 * @returns {boolean} 
	 */
	isRegionUnbindable: function (region, page) {
		// если нет модели или регион на другой странице
		if (!region || region.pageId() !== page.id()) {
			return true;
		}

		// всегда создавать если есть дети 
		// или это повторяющаяся группа табличного вида
		if (region.hasOwnProperty("regions") &&
			(region.regions().length > 0 ||
				region instanceof FC.BusinessModels.DocumentEditor.TableRegion && region.isRepeatableGroup)) {
			return false;
		}

		// если регион NotMatched или невидим
		if (!region.status() || !region.isVisible()) {
			return true;
		}

		// если у региона нет прямоугольников - нечего показывать
		var coords = region.coordinates();
		if (!coords || coords.length === 0) {
			return true;
		}

		return false;
	},

	/**
	 * Видим ли регион
	 * @param {} region модель региона
	 * @param {} page текущая страница
	 * @returns {boolean} 
	 */
	isRegionInvisible: function (region, page) {
		// если есть дети, но статус NotMatched или модель невидима - то невидим и регион
		if (region.hasOwnProperty("regions") && (!region.status() || !region.isVisible())) {
			return true;
		}

		// нет прямоугольников - не на что смотреть
		var coords = region.coordinates();
		if (!coords || coords.length === 0) {
			return true;
		}

		// если прямоугольник не один - точно видим
		if (coords.length !== 1) {
			return false;
		}

		// бывает случай, когда региону назначается один прямоугольник по размеру совпадающий со страницей
		// такие показывать не надо
		if (coords[0].x1() === 0 &&
			coords[0].y1() === 0 &&
			coords[0].x2() === page.width() &&
			coords[0].y2() === page.height()) {
			return true;
		}

		return false;
	},

	/**
	 * Создать вьюмодель региона
	 * @param {} region модель региона
	 * @param {} parentRegion вьюмодель родителя
	 * @param {} page текущая страница
	 * @param {} allRegions все регионы текущей страницы
	 * @returns {} вьюмодель региона или null
	 */
	createRegion: function (region, parentRegion, page, allRegions) {
		// если нельзя создавать вьюмодель - не создаём
		if (this.isRegionUnbindable(region, page)) {
			return null;
		}

		// определим видимость
		var isInvisible = this.isRegionInvisible(region, page);

		// глубина региона = глубина родителя+1
		var depth = parentRegion ? parentRegion.depth + 1 : 1;

		var regionVm = null;

		// если регион - табличный
		if (region instanceof FC.BusinessModels.DocumentEditor.TableRegion) {
			// может быть повторяющаяся группа с табличной разметкой
			if (region.isRepeatableGroup) {
				regionVm = new FC.Viewmodels.DocumentEditor.TableGroupRegionViewModel(region, parentRegion, page, depth);
			} else {
				regionVm = new FC.Viewmodels.DocumentEditor.TableRegionViewModel(region, parentRegion, page, depth);
			}
		}

		// если регион группы
		if (region instanceof FC.BusinessModels.DocumentEditor.GroupRegion) {
			regionVm = new FC.Viewmodels.DocumentEditor.GroupRegionViewModel(region, parentRegion, page, depth);
		}

		// если есть дочерние регионы - создадим и их
		var childRegions = region.regions();

		// если уже создали вьюмодель, то это регионы таблицы или групп
		if (regionVm) {
			regionVm.isVisible(!isInvisible);

			if (childRegions && childRegions.length > 0) {
				regionVm.updateRegions(childRegions, allRegions);
			}

			allRegions.push(regionVm);
			return regionVm;
		}

		// если родитель - таблица, то нужно создать ячейку
		if (parentRegion && parentRegion instanceof FC.Viewmodels.DocumentEditor.TableRegionViewModel) {
			if (parentRegion instanceof FC.Viewmodels.DocumentEditor.TableGroupRegionViewModel) {
				regionVm = new FC.Viewmodels.DocumentEditor.GroupCellRegionViewModel(region, parentRegion, page, depth);
			} else {
				regionVm = new FC.Viewmodels.DocumentEditor.CellRegionViewModel(region, parentRegion, page, depth);
			}
		} else {
			regionVm = new FC.Viewmodels.DocumentEditor.RegionViewModel(region, parentRegion, page, depth);
		}

		allRegions.push(regionVm);
		return regionVm;
	},

	updateRegion: function (regionViewModel, regionModel, allRegions) {
		allRegions.push(regionViewModel);

		// если есть дочерние регионы - обновим их
		var childRegions = regionModel.regions();

		if (childRegions && childRegions.length > 0) {
			regionViewModel.updateRegions(childRegions, allRegions);
		}
	},

	/**
	 * создать полнотекстовый регион
	 * @param {} region модель региона
	 * @param {} page текущая страница
	 * @returns {} 
	 */
	createTextRegion: function (region, page) {
		if (!region) {
			return null;
		}

		return new FC.Viewmodels.DocumentEditor.TextRegionViewModel(region, page);
	}
};