﻿/**
* Скрываемый элемент формы данных (скрываются его дети)
* @constructor
* @param {string} id - ИД элемента
* @param {number} offset - вертикальная позиция элемента (за вычетом всех скрываемых)
* @param {boolean} visible - видимость
* @param {} childrenObservable - observable-массив всех детей
* @param {} displayChildrenObservable 0 observable-массив показываемых детей
* @param {boolean} alwaysVisible - дети всегда видимы
*/
FC.Viewmodels.DocumentEditor.HideableItem = function (id, offset, visible, childrenObservable, displayChildrenObservable, alwaysVisible, isSectionGroup, field, parentHideableIds) {

	/** ИД элемента */
	this.id = ko.observable(id);

	this.isSectionGroup = isSectionGroup;

	this.parentHideableIds = parentHideableIds;

	/** высоты - видимая, невидимая сверху, невидимая снизу */
	this.heights = ko.observable({
		invisibleHeightUpper: 0,
		invisibleHeightLower: 0,
		visibleHeight: 0
	});

	this.field = field;

	/** вертикальная позиция элемента (за вычетом всех скрываемых) */
	this.offset = ko.observable(offset);

	/** видимость */
	this.visible = ko.observable(visible);

	/** дочерние элементы всегда видимы */
	this.allChildrenVisible = ko.observable(!!alwaysVisible);

	/**
	 * все дочерние элементы
	 */
	this.children = childrenObservable;

	/**
	 * видимые дочерние элементы
	 */
	this.displayChildren = displayChildrenObservable;

	var self = this;
	/**
	 * общая высота элемента
	 */
	this.totalHeight = ko.computed(function () {
		var heights = self.heights();
		return heights.invisibleHeightUpper + heights.invisibleHeightLower + heights.visibleHeight;
	});
};

/*
* обновить вертикальное положение элемента и видимость
*/
FC.Viewmodels.DocumentEditor.HideableItem.prototype.refreshOffset = function () {
	var offset = FC.Viewmodels.DocumentEditor.HideableItem.getHideableOffset(this.id(), this.isSectionGroup, this.parentHideableIds);
	this.visible(offset.visible);
	this.offset(offset.offset);
	this.allChildrenVisible(this.allChildrenVisible() || !!offset.isChildHideble);

	if (!offset.visible && (!this.allChildrenVisible() || this.children().length === 0)) {
		this.recalculateHeights(0);
	}
};

FC.Viewmodels.DocumentEditor.HideableItem.prototype.bindExpanders = function () {
	var fieldsPanel = window.batchLayout.batchLayoutViewModel.UserEditor.fieldsViewmodel;

	var currentHideable = $('#formViewer_body [data-hideableId="' + this.id() + '"]');
	if (!currentHideable || currentHideable.length === 0) {
		return;
	}

	if (currentHideable.is("input")) {
		// повторяющееся поле, не группа
		currentHideable = currentHideable.parent();
	}

	currentHideable
		.find(".sectionExpander")
		.off(".sectionExpantion")
		.on("click.sectionExpantion", fieldsPanel.onSectionExpanderClicked);

	fieldsPanel.createGroupExpanders(currentHideable);

	currentHideable
		.find(".groupExpander")
		.off(".groupExpantion")
		.on("click.groupExpantion", fieldsPanel.onGroupExpanderClicked);
};

/**
 * Пересчитать высоты
 * @param {number} hideablesUpperTotal -общая высота скрываемых элементов выше текущего
 * @param {number} scrollTop - начало видимого окна (по вертикали)
 * @param {number} containerHeight  - высота видимого окна.
 * @returns {} 
 */
FC.Viewmodels.DocumentEditor.HideableItem.prototype.recalculateHeights = function (hideablesUpperTotal, scrollTop, containerHeight) {
	if (arguments.length === 0) {
		var fieldsPanel = window.batchLayout.batchLayoutViewModel.UserEditor.fieldsViewmodel;

		scrollTop = fieldsPanel.formVerticalScroll.scrollTop();
		containerHeight = $(fieldsPanel.containerElement()).height();

		var currentOffset = this.offset();
		hideablesUpperTotal = fieldsPanel.hideables()
			.filter(function (h) { return h.visible() && h.offset() < currentOffset; })
			.map(function (h) { return h.totalHeight(); })
			.reduce(function (a, b) { return a + b; }, 0);
	}

	var children = this.children();

	if (!this.visible() && !this.allChildrenVisible() || children.length === 0 || this.field.collapseState === 0) {
		this.displayChildren([]);
		this.heights({
			invisibleHeightLower: 0,
			invisibleHeightUpper: 0,
			visibleHeight: 0
		});

		return;
	}

	this.refreshOffset();
	var groupTop = this.offset() + hideablesUpperTotal;

	var getChildElementHeight = function (parentField, childField) {
		var elem;
		if (childField.fieldType < 3) {
			elem = $("#" + childField.UniqueIdentifier);
		} else {
			elem = $("#" + childField.DomId);
		}

		if (elem) {
			return elem.outerHeight();
		}

		return 0;
	};

	var averageHeight = 26;
	if (!children[0].hasOwnProperty("elementHeight")) {
		this.displayChildren([]);
		this.displayChildren.push(children[0]);
		averageHeight = getChildElementHeight(this.field, this.displayChildren()[0]);
		this.displayChildren([]);
	}

	var groupHeight = 0;
	var supposedHeight = 0;
	var supposedTop = groupTop;
	var displayed = [];
	var invisibleUpper = 0;
	var invisibleLower = 0;
	var visibleHeight = 0;
	var i;
	var deltaY = 5;

	for (i = 0; i < children.length; i++) {
		if (!children[i].hasOwnProperty("elementHeight")) {
			children[i].elementHeight = averageHeight;
		}

		supposedHeight += children[i].elementHeight;
		var supposedBottom = groupTop + supposedHeight;

		// элемент выше окна
		if (scrollTop > supposedBottom + deltaY && !this.allChildrenVisible()) {
			groupHeight += children[i].elementHeight;

			invisibleUpper += children[i].elementHeight;
			supposedTop += children[i].elementHeight;
			continue;
		}

		// элемент в окне
		if (this.allChildrenVisible() || scrollTop <= supposedBottom + deltaY && (containerHeight + scrollTop) > supposedTop) {
			displayed.push(children[i]);
		}

		supposedTop += children[i].elementHeight;
	}

	var previousDisplayed = this.displayChildren();
	var needBindExtenders = false;
	this.displayChildren(displayed);
	for (i = 0; i < displayed.length; i++) {
		var processChildHideables = function (field) {
			if (field instanceof FC.Viewmodels.DocumentEditor.FieldsGroupAdapter ||
				field instanceof FC.Viewmodels.DocumentEditor.SectionAdapter) {
				var fieldChildren = field.children();
				if (fieldChildren.length > 0) {
					fieldChildren.forEach(processChildHideables);
					return true;
				} else {
					return false;
				}
			}

			if (field instanceof FC.Viewmodels.DocumentEditor.RepeatableAdapter ||
				field instanceof FC.Viewmodels.DocumentEditor.TableAdapter) {

				if (!field.hideable) {
					field.recalculateDisplayedChildren();
				}

				field.hideable.refreshOffset();

				field.hideable.recalculateHeights(hideablesUpperTotal + invisibleUpper, scrollTop, containerHeight);
				return true;
			}

			return false;
		};

		// нужно дёрнуть recalculateHeights у детей
		if (processChildHideables(displayed[i]) &&
			!needBindExtenders &&
			previousDisplayed.filter(function (pd) { return pd === displayed[i]; }).length === 0) {

			needBindExtenders = true;
		}

		displayed[i].elementHeight = getChildElementHeight(this.field, displayed[i]);

		groupHeight += displayed[i].elementHeight;
		visibleHeight += displayed[i].elementHeight;
	}

	if (needBindExtenders || this.isSectionGroup) {
		this.bindExpanders();
	}

	var invisibleLowerIndex = -1;
	if (displayed.length === 0) {
		if (invisibleUpper === 0) {
			invisibleLowerIndex = 0;
		}
	} else {
		invisibleLowerIndex = children.indexOf(displayed[displayed.length - 1]) + 1;
	}

	if (invisibleLowerIndex !== -1) {
		for (i = invisibleLowerIndex; i < children.length; i++) {
			groupHeight += children[i].elementHeight;
			invisibleLower += children[i].elementHeight;
		}
	}

	this.heights({
		invisibleHeightUpper: invisibleUpper,
		invisibleHeightLower: invisibleLower,
		visibleHeight: visibleHeight
	});

};

/**
 * Получить вертикальное положение скрываемого элемента по его id
 * @param {} hideableId - ид скрываемого элемента
 * @returns {} 
 */
FC.Viewmodels.DocumentEditor.HideableItem.getHideableOffset = function (hideableId, isSectionGroup, hideableParentIds) {
	var fieldsPanel = window.batchLayout.batchLayoutViewModel.UserEditor.fieldsViewmodel;

	var $hideables = $('#formViewer_body [data-hideableId]');
	var findHideableElement = function (id) {
		for (var k = 0; k < $hideables.length; k++) {
			if ($hideables[k].getAttribute("data-hideableId") === id) {
				return $($hideables[k]);
			}
		}

		return null;
	};

	var currentHideable = findHideableElement(hideableId);

	if (!currentHideable || currentHideable.length === 0) {
		return {
			visible: false, offset: 0
		};
	}

	if (currentHideable.is(":hidden") && currentHideable.is("tbody")) {
		currentHideable = currentHideable.prev();
	}

	if (!isSectionGroup && currentHideable.is(":hidden")) {
		return {
			visible: false, offset: 0
		};
	}

	var containerOffset = currentHideable.offset().top;
	if (currentHideable.is("thead")) {
		containerOffset += currentHideable.outerHeight();
	}

	var actualScrollTop = $(fieldsPanel.scrollableElement()).offset().top;

	var i;
	var containerTop = containerOffset - actualScrollTop;
	var offsetOnForm = containerTop;
	var hideables = fieldsPanel.hideables()
		.filter(function (h) { return h.visible() || h.allChildrenVisible(); })
		.sort(function (a, b) { return a.offset() - b.offset(); });

	for (i = 0; i < hideables.length; i++) {
		// текущий нам не нужен
		if (hideables[i].id() === hideableId) {
			continue;
		}

		// нам интересны только те hideables, которые находятся или в корне, или на одном из родительских уровней текущего
		if (hideables[i].parentHideableIds.length > 0 &&
			hideableParentIds.indexOf(hideables[i].parentHideableIds[hideables[i].parentHideableIds.length - 1]) < 0) {
			continue;
		}

		// родителей текущего элемента не вычитаем
		if (hideableParentIds.indexOf(hideables[i].id()) > -1) {
			continue;
		}

		if (hideables[i].offset() > 0 && containerTop > hideables[i].offset()) {
			var hideable = findHideableElement(hideables[i].id());

			if (!hideable) {
				// нет в разметке
				continue;
			}
			var hOffset = hideable.offset().top - actualScrollTop;
			if (hideable.is("thead")) {
				hOffset += hideable.outerHeight();
			}

			if (offsetOnForm < hOffset) {
				continue;
			}

			var childHeight = 0;
			hideable.children(":not(.instance_menu-popup-container):not(.documentEditor_addItem)").each(function () { childHeight += $(this).height(); });
			containerTop -= childHeight;
		}
	}

	if (containerTop < 0) {
		return {
			visible: false, offset: 0
		};
	}

	return {
		visible: true, offset: containerTop, isChildHideble: hideableParentIds.length > 0
	};
};

/**
 * Получить или создать скрываемый элемент
 * @param {} hideableField - вьюмодель поля со скрываемыми детьми
 * @returns {FC.Viewmodels.DocumentEditor.HideableItem} 
 */
FC.Viewmodels.DocumentEditor.HideableItem.getOrCreateHideable = function (hideableField) {
	var fieldsPanel = window.batchLayout.batchLayoutViewModel.UserEditor.fieldsViewmodel;
	var hideables = fieldsPanel.hideables();

	var hideableId = hideableField instanceof FC.Viewmodels.DocumentEditor.TableAdapter ? hideableField.bodyDomId : hideableField.UniqueIdentifier;
	var childrenObservable = hideableField instanceof FC.Viewmodels.DocumentEditor.TableAdapter ? hideableField.rows : hideableField.children;
	var displayedChildrenObservable = hideableField instanceof FC.Viewmodels.DocumentEditor.TableAdapter ? hideableField.displayedRows : hideableField.displayedChildren;
	var isSectionGroup = hideableField.model instanceof FC.BusinessModels.DocumentEditor.SectionsGroup;

	var currentHideable = hideables.filter(function (h) { return h.id() === hideableId; });
	if (currentHideable.length > 0) {
		return currentHideable[0];
	}

	var hideableParent = hideableField.parent;
	var hideableParentIds = [];
	while (hideableParent) {
		if (hideableParent.hasOwnProperty("hideable")) {
			hideableParentIds.unshift(hideableParent instanceof FC.Viewmodels.DocumentEditor.TableAdapter ? hideableParent.bodyDomId : hideableParent.UniqueIdentifier);
		}
		hideableParent = hideableParent.parent;
	}

	var hideableOffsetInfo = FC.Viewmodels.DocumentEditor.HideableItem.getHideableOffset(hideableId, isSectionGroup, hideableParentIds);

	var newInfo = new FC.Viewmodels.DocumentEditor.HideableItem(
		hideableId,
		hideableOffsetInfo.offset,
		hideableOffsetInfo.visible,
		childrenObservable,
		displayedChildrenObservable,
		isSectionGroup || hideableOffsetInfo.isChildHideble,
		isSectionGroup,
		hideableField,
		hideableParentIds);

	for (var i = 0; i < hideables.length; i++) {
		if (hideables[i].offset() > hideableOffsetInfo.offset) {
			hideables.splice(i, 0, newInfo);
			return newInfo;
		}
	}

	hideables.push(newInfo);
	return newInfo;
};

