﻿/**
 * Меню для управления повторяющимися элементами на форме данных
 * @param {} id идентификатор меню
 * @param {} addItemOnIndex функция добавления элемента на позиции
 * @param {} removeSelectedItems удалить выбранные элементы по их индексам
 * @param {} removeAllItems удалить все элементы
 * @param {} getItemIndex получить индекс элемента
 * @param {} selectedField выбранное поле
 * @param {} countInstances получить число элементов
 * @param {} scrollToInstance подскроллить форму данных к элементу по его индексу
 * @returns {} 
 */
FC.Viewmodels.DocumentEditor.InstanceMenu = function (id, addItemOnIndex, removeSelectedItems, removeAllItems,
	getItemIndex, selectedField, countInstances, scrollToInstance, onSelectedInstancesChanged, customActions) {
	var self = this;

	self.customActions = customActions;
	// индексы выбранных инстансов
	self.selectedInstances = ko.observableArray();

	// видно ли меню управления инстансами
	self.instanceMenuVisible = ko.observable(false);

	// положение меню управления - верхний левый угол
	self.instanceMenuTop = ko.observable(0);
	self.instanceMenuLeft = ko.observable(0);

	// идентификатор меню в DOM
	self.instanceMenuPopupDomId = "instanceMenu_" + id;

	self.selectedField = selectedField;
	self.countInstances = countInstances;

	var fieldsPanel = window.batchLayout.batchLayoutViewModel.UserEditor.fieldsViewmodel;
	fieldsPanel.disposables.push(self.selectedField.subscribe(function (value) {
		// нужно сбрасывать выделение инстансов, если выделили поле
		if (value && self.selectedInstances().length > 0) {
			self.selectedInstances([]);
		}
	}));

	fieldsPanel.disposables.push(self.selectedInstances.subscribe(function (values) {
		if (onSelectedInstancesChanged) onSelectedInstancesChanged(values);
		if (!values || values.length === 0) {
			// если нет выделенных инстансов - выходим из режима
			self.keyboardEventAttached = false;
			$(window).off("keydown.instances");
			fieldsPanel.selectingInstances(false);
			self.firstSelectedInstance(-1);
			
		}
	}));

	/**
	 * Добавить инстанс на позиции
	 * @param {} index позиция
	 */
	self.addRepeatedItemOnIndex = function (index) {
		addItemOnIndex(index);
	};

	/**
	 * Добавить инстанс выше выделенного 
	 */
	self.addInstanceAbove = function () {
		if (self.selectedInstances().length === 0 || self.selectedInstances().length > 1) {
			return;
		}

		addItemOnIndex(self.selectedInstances()[0]);
	};

	/**
	 * Добавить инстанс ниже выделенного 
	 */
	self.addInstanceBelow = function () {
		if (self.selectedInstances().length === 0 || self.selectedInstances().length > 1) {
			return;
		}

		addItemOnIndex(self.selectedInstances()[0] + 1);
	};

	/**
	 * Удалить выбранные инстансы 
	 */
	self.removeSelectedInstances = function () {
		var instances = self.selectedInstances();

		if (instances.length === 0) {
			return;
		}

		instances.sort(function (i1, i2) { return i1 - i2; });
		var topInstance = instances[0];
		removeSelectedItems(instances);

		self.selectedInstances([]);

		// подскроллимся к инстансу, который был перед самым верхним удалённым
		setTimeout(scrollToInstance(topInstance - 1), 100);
	};

	/**
	 * Удалить все инстансы 
	 */
	self.removeAllInstances = function () {
		removeAllItems();

		self.selectedInstances([]);

		setTimeout(scrollToInstance(-1), 100);
	};

	/**
	 * Стиль меню инстансов
	 */
	self.instanceMenuStyle = ko.computed(function () {
		return {
			top: self.instanceMenuTop() + "px",
			left: self.instanceMenuLeft() + "px"
		};
	});

	/**
	 * Скрыть меню инстансов 
	 */
	self.hideInstancesMenu = function () {
		$(window).off("click.instanceMenu");
		$(window).off("contextmenu.instanceMenu");
		$("#formViewer_body").off("wheel.instanceMenu");

		self.instanceMenuVisible(false);
	};

	/**
	 * Показать меню инстансов 
	 */
	self.showInstanceMenu = function (instance, event) {
		$(window).on("click.instanceMenu", self.hideInstancesMenu);
		$(window).on("contextmenu.instanceMenu", self.hideInstancesMenu);
		$("#formViewer_body").on("wheel.instanceMenu", self.hideInstancesMenu);

		self.instanceMenuVisible(true);
		var menuPopup = $("#" + self.instanceMenuPopupDomId);
		var popupHeight = menuPopup.outerHeight();

		var instanceTop = event.clientY;

		if (window.innerHeight < event.clientY + popupHeight) {
			instanceTop = instanceTop - popupHeight;
		}

		self.instanceMenuTop(instanceTop);
		self.instanceMenuLeft(event.clientX);

		event.originalEvent.cancelBubble = true;
		if (event.originalEvent.stopPropagation) {
			event.originalEvent.stopPropagation();
		}
	};

	/**
	 * Сбросить выделение инстансов
	 */
	self.unselectInstances = function () {
		$("#formViewer_body").off("mousedown.instances");
		if (self.selectedInstances() && self.selectedInstances().length > 0) {
			self.selectedInstances([]);
		}
		self.hideInstancesMenu();
	};

	self.unselectInstance = function (index) {
		
		if (self.selectedInstances() && self.selectedInstances().indexOf(index) >= 0) {
			self.selectedInstances().splice(self.selectedInstances().indexOf(index),1);
		}
		self.hideInstancesMenu();
	};

	// режим выделения строк с прижатой мышью
	self.selectionModeOn = ko.observable(false);

	/**
	 * Добавить инстанс к выбранным
	 * @param {} instanceIndex индекс нового инстанса
	 */
	self.addInstanceToSelection = function (instanceIndex) {
		self.lastSelectedInstance(instanceIndex);

		var firstSelected = self.firstSelectedInstance();
		var firstIndex = Math.min(firstSelected, instanceIndex);
		var lastIndex = Math.max(firstSelected, instanceIndex);

		var instances = [];
		for (var i = firstIndex; i <= lastIndex; i++) {
			instances.push(i);
		}

		self.selectedInstances(instances);
	};

	/**
	 * Закончить выделять инстансы по нажатию мыши
	 */
	self.stopSelecting = function () {
		$(window).off("mouseup.instances");
		$(document.body).off("mouseleave.instances");
		self.selectionModeOn(false);
	};

	/**
	 * В режиме выделения с прижатой мышью
	 * выделяем инстанс, если мы на него навели мышь
	 */
	self.instanceMenuEnter = function (instance, event) {
		if (!self.selectionModeOn()) {
			return;
		}

		var instanceIndex = getItemIndex(instance);
		if (event.ctrlKey) {
			if (self.selectedInstances().indexOf(instanceIndex) < 0) {
				self.selectedInstances.push(instanceIndex);
			}
			return;
		}
		self.addInstanceToSelection(instanceIndex);
	};

	// индекс последнего выделенного инстанса
	self.lastSelectedInstance = ko.observable(-1);

	/**
	 * Обработка горячих клавиш
	 * @param {any} event
	 */
	var keyDown = function (event) {
		var eventProcessed = function () {
			event.preventDefault();

			if (event.originalEvent.stopPropagation) {
				event.originalEvent.stopPropagation();
			}

			event.originalEvent.cancelBubble = true;
		};

		// обработать нажатие стрелок
		var processArrows = function (calculateCurrentInstance) {
			var lastInstance = self.lastSelectedInstance();
			var currentInstance = calculateCurrentInstance(lastInstance);

			if (currentInstance >= 0) {
				if (event.ctrlKey || event.shiftKey) {
					self.addInstanceToSelection(currentInstance);
				} else {
					self.lastSelectedInstance(currentInstance);
					self.selectedInstances([currentInstance]);
					self.firstSelectedInstance(currentInstance);
				}

				scrollToInstance(currentInstance);
			}

			eventProcessed();
		}


		switch (event.keyCode) {
			case 38: // up
				processArrows(function (lastInstance) {
					return lastInstance - 1;
				});
				break;
			case 40: // down
				processArrows(function (lastInstance) {
					return lastInstance + 1 < self.countInstances() ? lastInstance + 1 : -1;
				});
				break;
			case 45: // insert
				if (event.shiftKey) {
					self.addInstanceBelow();
				} else {
					self.addInstanceAbove();
				}
				eventProcessed();
				break;
			case 46: // delete
				self.removeSelectedInstances();
				eventProcessed();
				break;
		}
	};

	// подписаны ли на события клавиатуры
	self.keyboardEventAttached = false;

	// первый выделенный инстанс
	self.firstSelectedInstance = ko.observable(-1);

	// выделить инстанс
	self.selectInstance = function (instance, event) {
		// если было выделение у других меню их нужно сбросить
		if (fieldsPanel.lastInstanceMenu && fieldsPanel.lastInstanceMenu !== self) {
			fieldsPanel.lastInstanceMenu.unselectInstances();
		}

		fieldsPanel.lastInstanceMenu = self;

		// по левой кнопке прячем меню и входим в режим выделения-с-прижатой-мышью
		if (event && event.button !== 2) {
			self.hideInstancesMenu();
			self.selectionModeOn(true);
			$(window).on("mouseup.instances", self.stopSelecting);
			$(document.body).on("mouseleave.instances", self.stopSelecting);

			if (event && event.originalEvent.stopPropagation) {
				event.originalEvent.stopPropagation();
			}
		}

		var instanceIndex = getItemIndex(instance);

		// подписываемся на события клавиатуры
		if (!self.keyboardEventAttached) {
			$(window).on("keydown.instances", keyDown);
			fieldsPanel.selectingInstances(true);
			self.firstSelectedInstance(instanceIndex);
			self.keyboardEventAttached = true;
		}

		// запоминаем на какой инстанс нажали первым
		if (!event || !event.shiftKey || self.firstSelectedInstance() < 0) {
			self.firstSelectedInstance(instanceIndex);
		}

		if (event) event.originalEvent.cancelBubble = true;

		$("#formViewer_body").on("mousedown.instances", self.unselectInstances);

		// сбрасываем выделение поля - нельзя находиться в двух режимах одновременно
		self.selectedField(null);

		if (event && event.shiftKey) {
			self.addInstanceToSelection(instanceIndex);
			return;
		}

		if (event && event.button === 2 && self.selectedInstances().indexOf(instanceIndex) >= 0) {
			return;
		}

		self.lastSelectedInstance(instanceIndex);

		if (event && event.ctrlKey) {
			if (self.selectedInstances().indexOf(instanceIndex) >= 0) {
				self.selectedInstances.splice(self.selectedInstances().indexOf(instanceIndex), 1);
			} else {
				self.selectedInstances.push(instanceIndex);
			}

			return;
		}
		if (self.selectedInstances().length !== 1 || self.selectedInstances()[0] !== instanceIndex) {
			self.selectedInstances([instanceIndex]);
		}
	};
};