﻿var FC = FC || {};
FC.Viewmodels = FC.Viewmodels || {};
FC.Viewmodels.ImageEditor = FC.Viewmodels.ImageEditor || {};

/**
 * Рисователь на канвасе 
 * @param {} pageDrawer рисователь из DocumentEditor (FC.DefaultDrawer)
 * @returns {} 
 */
FC.Viewmodels.ImageEditor.CanvasLayer = function (pageDrawer) {

	this.pageDrawer = pageDrawer;

	/**
	 * Нарисовать заполненный прямоугольник (создать команду)
	 * @param {} rectangle прямоугольник типа FC.Viewmodels.ImageEditor.Rectangle или FC.BusinessModels.DocumentEditor.Coordinates
	 * @param {} color цвет 
	 * @returns {} 
	 */
	this.drawFilledRectangle = function (rectangle, color) {
		this.pageDrawer.addDrawRectangleCommand({
			rectangle: {
				left: rectangle.x1(),
				top: rectangle.y1(),
				width: rectangle.x2() - rectangle.x1(),
				height: rectangle.y2() - rectangle.y1()
			},
			fill: {
				color: color
			}
		});

	};

	/**
	 * Нарисовать прямоугольник (создать команду)
	 * @param {} rectangle  прямоугольник типа FC.Viewmodels.ImageEditor.Rectangle или FC.BusinessModels.DocumentEditor.Coordinates
	 * @param {} width ширина контура
	 * @param {} color цвет контура
	 * @param {} backColor цвет фона
	 * @returns {} 
	 */
	this.drawRectangle = function (rectangle, width, color, backColor, dashes) {
		var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();

		var zoom = docEditor.zoom();
		var stroke = width ? { color: color, thickness: width / zoom } : null;
		var fill = backColor ? { color: backColor } : null;

		dashes = dashes || [];
		var convertedDashes = dashes.map(function (d) { return d / zoom; });
		if (stroke && convertedDashes.length > 0) {
			stroke.dash = convertedDashes;
		}

		this.pageDrawer.addDrawRectangleCommand({
			rectangle: {
				left: rectangle.x1(),
				top: rectangle.y1(),
				width: rectangle.x2() - rectangle.x1(),
				height: rectangle.y2() - rectangle.y1()
			},
			fill: fill,
			stroke: stroke
		});
	};

	/**
	 * Нарисовать полигон с единой границей из массива прямоугольноков (создать команду)
	 * @param {} rectangles прямоугольники типа FC.Viewmodels.ImageEditor.Rectangle или FC.BusinessModels.DocumentEditor.Coordinates
	 * @param {} width ширина контура
	 * @param {} color цвет контура
	 * @param {} backColor цвет фона
	 * @returns {} 
	 */
	this.drawPolygon = function (rectangles, width, color, backColor) {
		var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();

		var stroke = width ? { color: color, thickness: width / docEditor.zoom() } : null;
		var fill = backColor ? { color: backColor } : null;
		var points = FC.Utils.polygonCommon.getPolygonPoints(ko.toJS(rectangles));

		if (points) {
			this.pageDrawer.addDrawPolygonCommand({
				points: points,
				fill: fill,
				stroke: stroke
			});
		}
	};

	/**
	 * Нарисовать линию (создать команду)
	 * @param {} point1 первая точка
	 * @param {} point2 вторая точка
	 * @param {} width ширина линии
	 * @param {} color цвет
	 * @returns {} 
	 */
	this.drawLine = function (point1, point2, width, color, dashes) {
		if (!width) {
			return;
		}

		var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();
		var zoom = docEditor.zoom();

		dashes = dashes || [];
		var convertedDashes = dashes.map(function (d) { return d / zoom; });

		this.pageDrawer.addDrawLineCommand({
			color: color,
			thickness: width / zoom,
			x1: point1.x,
			x2: point2.x,
			y1: point1.y,
			y2: point2.y,
			dashes: convertedDashes
		});
	};

	/**
	 * Нарисовать кружок с тенью (создать команду)
	 * @param {} point центр круга - точка
	 * @param {} radius радиус круга
	 * @param {} bgColor цвет фона
	 * @param {} shadowColor цвет тени
	 * @param {} shadowOffsetX смещение тени по x
	 * @param {} shadowOffsetY смещение тени по y
	 * @param {} shadowBlur размытие тени
	 * @returns {} 
	 */
	this.drawCircleWithShadow = function (point, radius, bgColor, shadowColor, shadowOffsetX, shadowOffsetY, shadowBlur) {
		this.pageDrawer.addDrawCircleCommand({
			backgroundColor: bgColor,
			shadowColor: shadowColor,
			point: point,
			radius: radius,
			shadowOffsetX: shadowOffsetX,
			shadowOffsetY: shadowOffsetY,
			shadowBlur: shadowBlur
		});
	};

	var getFontDescription = function (fontSize, fontFace, zoom) {
		return Math.round(fontSize / zoom) + "px " + fontFace;
	};

	/**
	 * Написать на канвасе текст (создать команду)
	 * @param {} point левая нижняя точка
	 * @param {} text текст
	 * @param {} color цвет
	 * @param {} fontSize размер шрифта
	 * @param {} fontFace название шрифта
	 * @returns {} 
	 */
	this.drawText = function (point, text, color, fontSize, fontFace) {
		var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();
		var zoom = docEditor.zoom();

		this.pageDrawer.addDrawTextCommand(
			{
				text: text,
				point: point,
				color: color,
				font: getFontDescription(fontSize, fontFace, zoom)
			});
	};

	/**
	 * Получить ширину текста
	 * @param {} text текст
	 * @param {} fontSize размер шрифта
	 * @param {} fontFace шрифт
	 * @returns {} 
	 */
	this.measureText = function (text, fontSize, fontFace) {
		var docEditor = window.batchLayout.batchLayoutViewModel.DocumentEditor();
		return docEditor.measureTextWidth(text, fontSize, fontFace);
	};
};