﻿FC.Viewmodels.Validation = FC.Viewmodels.Validation || {};
/**
 * Валидатор для денежной суммы
 */
FC.Viewmodels.Validation.CurrencyValidator = function currencyValidator(serverValidator) {

	var self = this;
	self.serverValidator = serverValidator;

	/**
	 * Валидоровать денежную сумму.
	 * возвращает null если текст является денежной суммой
	 * или объект с полем error, в котором содержится текст ошибки, если не денежная сумма.
	 * Денежная сумма - это число, у которого спереди или сзади может быть символ валюты.
	 * Символ валюты от числа может отделять пробел.
	 * Критерии числа - см. FC.Viewmodels.Validation.NumberValidator.
	 * Возможные символы валюты приходят с сервера в поле self.serverValidator.PossibleCurrencySymbols
	 */
	self.Validate = function (text) {

		// проверка на пустой текст - у всех валидаторов общая, здесь не проверяем
		// если серверный валидатор другого типа, тоже не проверяем
		if (!text || !self.serverValidator || self.serverValidator.ValidatorType !== 'CurrencyValidator') {
			// не можем проверить - значит не должны и падать
			return null;
		}

		// создаём валидатор числа
		var numberValidator = new FC.Viewmodels.Validation.NumberValidator(self.serverValidator, true);

		// отрицательность может задаваться минусом или скобками. Проверим скобки
		var isNegative = false;
		if (self.serverValidator.CanHaveParentheses) {
			var removalResult = numberValidator.removeParentheses(text);
			isNegative = removalResult.isNegative;
			text = removalResult.text;
		}

		// если скобок не было - проверим минус
		if (!isNegative) {
			var minusRemovalResult = numberValidator.removeMinus(text);
			if (!minusRemovalResult) {
				// много минусов, например
				return { error: FC.Environment.Localization.GlobalResources.FormatError };
			}

			text = minusRemovalResult.text;
			isNegative = minusRemovalResult.isNegative;
		}
		
		// ищем символы валюты и разделяем их от остального
		var currencyResult = self._parseCurrencySymbols(text);

		// может вернуться или ошибка, или оставшиеся после всех вырезаний символы
		// если ошибка - возвращаем её
		if (currencyResult.hasOwnProperty('error') || !currencyResult.hasOwnProperty('reminder')) {
			return currencyResult;
		}		

		// проверяем остаток на то, какое там число
		return numberValidator.Validate(currencyResult.reminder, isNegative);
	};

	/**
	 * Вырезать из текста символ валюты, если от там есть
	 */
	self._parseCurrencySymbols = function (text) {

		// если с сервера не пришли символы валюты, ничего не вырезаем
		if (!self.serverValidator.PossibleCurrencySymbols || self.serverValidator.PossibleCurrencySymbols.length === 0) {
			return { reminder: text };
		}

		var currencySymbol = null;

		// ищем с точностью до регистра
		text = text.trim().toLowerCase();

		// сортируем по убыванию размера разделителей - тогда сначала мы ищем длинные, а потом короткие
		// (например, чтобы различить руб. и р)
		var currencySymbols = self.serverValidator.PossibleCurrencySymbols.sort(function (a, b) {
			return b.length-a.length;
		});

		// ищем первый подходящий символ валюты
		for (var i = 0; i < currencySymbols.length; i++) {
			if (text.indexOf(currencySymbols[i].toLowerCase()) > -1) {
				currencySymbol = currencySymbols[i].toLowerCase();
				break;
			}
		}

		// если ничего не нашли - возвращаем текст неизменным
		if (!currencySymbol) {
			return { reminder: text };
		}

		// удалим пробелы с боков
		text = text.trim();

		// вырезаем символ валюты или сначала, или с конца
		var symbolRemoved = false;

		// сначала проверяем вначале
		if (text.indexOf(currencySymbol) === 0) {
			text = text.substring(currencySymbol.length).trim();
			symbolRemoved = true;
		}

		// потом проверяем в конце
		if (!symbolRemoved && text.indexOf(currencySymbol) === (text.length - currencySymbol.length)) {
			text = text.substring(0, text.length - currencySymbol.length).trim();
			symbolRemoved = true;
		}

		// если и после этого остались символы валюты - значит ошибка
		if (text.indexOf(currencySymbol) > -1) {
			if (symbolRemoved) {
				return { error: FC.Environment.Localization.GlobalResources.FormatError_TooManyCurrencySymbols };
			} else {
				return { error: FC.Environment.Localization.GlobalResources.FormatError_CurrencySymbolNotInBegginingOrEnd };
			}
		}

		return { reminder: text };
	};
};