﻿FC.Viewmodels.Validation = FC.Viewmodels.Validation || {};
/**
 * Валидатор для числа.
 * serverValidator - серверные настройки валидации
 * cantHavePercent - в числе не может быть процента (если это денежная сумма, например)
 */
FC.Viewmodels.Validation.NumberValidator = function numberValidator(serverValidator, cantHavePercent) {
	var self = this;

	self.serverValidator = serverValidator;

	/**
	 * Валидоровать число.
	 * возвращает null если текст является числом
	 * или объект с полем error, в котором содержится текст ошибки, если не число.
	 * Число может быть со скобками - тогда оно отрицательное (настраивается на сервере).
	 * Может быть минус спереди или сзади (второе настраивается на сервере).
	 * Так же есть различные разделители групп и дробной части (настраиваются на сервере)
	 * и интервал, которому число должно соответствовать.
	 */
	self.Validate = function (text, isNegative) {
		var sourceText = text;

		// Настройки валидации могут быть или для числа, или для денежной суммы
		// Иначе не можем валидировать
		if (!text ||
			!self.serverValidator ||
			(self.serverValidator.ValidatorType !== 'NumberValidator' &&
				self.serverValidator.ValidatorType !== 'CurrencyValidator')) {
			// не можем проверить - значит не должны и падать
			return null;
		}

		// удаляем скобки, если они могут быть
		if (!isNegative && self.serverValidator.CanHaveParentheses) {
			var removalResult = self.removeParentheses(text);
			isNegative = removalResult.isNegative;
			text = removalResult.text;
		}

		// удаляем минус, если есть
		if (!isNegative) {
			var minusRemovalResult = self.removeMinus(text);
			if (!minusRemovalResult) {
				// много минусов, например
				return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
			}

			isNegative = minusRemovalResult.isNegative;
			text = minusRemovalResult.text;
		}

		// если в числе может быть процент и он есть, и он в конце - удаляем
		if (!cantHavePercent) {
			if (text.indexOf('%') === text.length - 1) {
				text = text.substr(0, text.length - 1).trim();
			}
		}

		// проверяем на непозволительные символы
		if (!self._checkNumberSymbols(text)) {
			return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
		}

		var number;

		// если целое число - проверяем его на коррекность
		if (self.serverValidator.IsInteger) {
			// теперь у нас остались только цифры и разделители групп		
			var grouplessInteger = self._removeGroupSeparators(text, false);

			if (grouplessInteger.hasOwnProperty('error')) {
				return grouplessInteger;
			}

			if (!grouplessInteger.text.match(/^\d*$/)) {
				return { error: FC.Environment.Localization.GlobalResources.FormatError_NotInteger };
			}

			number = { wholePart: grouplessInteger.text, fractionalPart: '0' };
		} else {

			// разделяем на целую и дробную части, проверяем дробную часть
			number = self._parseAsFloatObjectPart(text);
			if (number.hasOwnProperty('error')) {

				// ошибка в дробной части
				return number;
			}
		}

		// теперь у нас остались только цифры и разделители групп	
		var groupRemovalResult = self._removeGroupSeparators(number.wholePart, false);

		if (groupRemovalResult.hasOwnProperty('error')) {
			// сепараторы неправильно расположены или разные
			return groupRemovalResult;
		}

		number.wholePart = groupRemovalResult.text;

		// если не нужно проверять интервал - то всё ок.
		if (!self.serverValidator.CheckInterval) {
			return null;
		}

		// собираем число из составляющих
		var floatingNumber = (isNegative ? '-' : '') + number.wholePart + '.' + number.fractionalPart;

		// проверяем интервал
		if (!self._checkNumberInterval(floatingNumber)) {

			// генерируем интервальную ошибку
			var intervalError = FC.Environment.Localization.GlobalResources.FormatErrorNumberRangeError
				.replace('%0', sourceText)
				.replace('%1', self.serverValidator.MinValue)
				.replace('%2', self.serverValidator.MaxValue);
			return { error: intervalError };
		}

		// число прошло все проверки
		return null;
	};

	/**
	 * Удалить минус из числа
	 */
	self.removeMinus = function (text) {
		if (!text) {
			return null;
		}

		text = text.trim();

		// если минус в конце
		if (text.indexOf('-') === text.length - 1) {

			// если минус может быть в конце - вырезаем и возвращаем что осталось
			if (self.serverValidator.CanHaveMinusOnRight) {
				return { text: text.substr(0, text.length - 1), isNegative: true };
			}

			// если не может - ошибка
			return null;
		}

		// если минус в начале - вырезаем
		if (text.indexOf('-') === 0) {
			return { text: text.substr(1), isNegative: true };
		}

		// нет минусов в разрешённых местах
		return { text: text, isNegative: false };
	};

	/**
	 * Проверить, что в строке только разрешённые символы
	 */
	self._checkNumberSymbols = function (text) {
		// могут быть только цифры и сепараторы групп и дробной части
		// прежде, чем разбирать на части и проверять каждую часть, проверим, что
		// там нет других символов. Если есть - однозначно отлуп
		var regexParts;
		if (self.serverValidator.IsInteger) {
			regexParts = self._distinctRegexParts(['\\d'], self.serverValidator.GroupSeparators);
		} else {
			regexParts = self._distinctRegexParts(['\\d'], self.serverValidator.GroupSeparators, self.serverValidator.DecimalSeparators);
		}

		// проверим по регулярному выражению
		var matchingRegex = self._createMatchingRegex('^', regexParts, '+$');
		return text.match(matchingRegex);
	};

	/**
	 * Проверить, что число попадает в интрервал
	 */
	self._checkNumberInterval = function (text) {
		var number = parseFloat(text);
		var min = parseFloat(self.serverValidator.MinValue);
		var max = parseFloat(self.serverValidator.MaxValue);

		return number >= min && number <= max;
	};

	/**
	 * Распарсить строку в число с плавающей точкой
	 */
	self._parseAsFloatObjectPart = function (text) {
		var fractionalPart = '0';

		// если дробная часть не обязательна
		if (!self.serverValidator.FractionalPartObligatory) {

			// проверим частный случай, что дробной части  нет
			var removalResult = self._removeGroupSeparators(text, true);
			if (!removalResult.hasOwnProperty('error')) {
				return { wholePart: removalResult.text, fractionalPart: fractionalPart };
			}

			if (removalResult.hasOwnProperty('floatAmbiguity')) {
				return removalResult;
			}
		}

		// сделаем регулярное выражение по которому можно найти разделитель дробной части
		var decimalPartFinder = self._createMatchingRegex('', self.serverValidator.DecimalSeparators, '\\d*$');
		var decimalPart = decimalPartFinder.exec(text);

		if (!decimalPart && self.serverValidator.FractionalPartObligatory) {
			// не нашли дробной части
			return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
		}

		// разрезаем строку на целую и дробную части
		if (decimalPart) {
			var decimalSymbolIndex = decimalPart.index;
			var decimalSymbol = decimalPart[0][0];
			fractionalPart = decimalPart[0].substr(1, decimalPart[0].length - 1);

			if (!self.serverValidator.FractionalPartCanBeLongerThanTwoDigits && fractionalPart.length > 2) {
				// слишком длинная дробная часть
				return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
			}

			text = text.substr(0, decimalSymbolIndex);
			if (text.indexOf(decimalSymbol) > -1) {
				// не может быть больше одного разделителя дробной части
				return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
			}
		}

		return { wholePart: text, fractionalPart: fractionalPart };
	};

	/**
	 * Вырезать из строки разделители групп
	 */
	self._removeGroupSeparators = function (text, canBeFloat) {

		// размер группы
		var groupSize = 3;

		if (!text) {
			return { text: text };
		}

		// если сепараторов нет - всё ок
		if (text.match(/^\d*$/)) {
			return { text: text };
		}

		if (text.length < groupSize + 1) {
			// у нас есть сепараторы не на своём месте
			return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
		}

		// проверим, нет ли неоднозначной ситуации, когда у нас есть две группы, но вторая может быть
		// ещё и дробной частью. Тогда должна вылезти ошибка. (напр. 123,456)
		if (canBeFloat && text.length <= groupSize * 2 + 1 && self.serverValidator.FractionalPartCanBeLongerThanTwoDigits) {
			var charInQuestion = text[text.length - 1 - groupSize];
			if (!self.serverValidator.PrefferedGroupSeparator ||
				self.serverValidator.PrefferedDecimalSeparator == self.serverValidator.PrefferedGroupSeparator ||
				charInQuestion != self.serverValidator.PrefferedGroupSeparator[0]) {
				if (self.serverValidator.GroupSeparators.indexOf(charInQuestion) > -1 &&
					self.serverValidator.DecimalSeparators.indexOf(charInQuestion) > -1) {

					return {
						error: FC.Environment.Localization.GlobalResources.FormatErrorIntegerOrReal,
						floatAmbiguity: true
					};
				}
			}
		}

		var groupRegex = self._createMatchingRegex(
			'^[0-9]{1,' + groupSize + '}(', //begin
			self.serverValidator.GroupSeparators, // values
			'[0-9]{' + groupSize + '})*$'); // end

		// проверяем положение групп по регулярному выражению
		if (!text.match(groupRegex)) {
			return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
		}

		// проверим, что все сепараторы одинаковые
		var separator = text[text.length - groupSize - 1];
		for (var i = text.length - groupSize - 1; i >= 0; i = i - (groupSize + 1)) {
			if (text[i] !== separator) {
				// не могут быть разные сепараторы
				return { error: FC.Environment.Localization.GlobalResources.FormatErrorNumberTypeMismatch };
			}
		}

		// нужно для сборки регулярного выражения - эскапируем точку
		if (separator === '.') {
			separator = '\\.';
		}

		// вырезаем сепараторы
		return { text: text.replace(new RegExp(separator, 'g'), '') };
	};

	/**
	 * Вырезать из строки скобки
	 * Скобки - признак отрицательного числа в неокторых культурах
	 */
	self.removeParentheses = function (text) {
		if (!text) {
			return { text: text, isNegative: false };
		}

		text = text.trim();
		if (text.length > 1 && text.indexOf('(') === 0 && text.indexOf(')') === text.length - 1) {
			return { text: text.substring(1, text.length - 1), isNegative: true };
		}

		return { text: text, isNegative: false };
	};

	// Эскапировать разделители, чтобы собрать потом из них регулярное выражение
	self._escapeSplitters = function (splitterArray) {
		if (!splitterArray || !splitterArray.hasOwnProperty('length') || splitterArray.length === 0) {
			return '';
		}

		// нужно эскапировать вот эти символы
		var escapableChars = '.$()*+?\\/-';

		var result = [];
		for (var i = 0; i < splitterArray.length; i++) {
			if (escapableChars.indexOf(splitterArray[i]) > -1) {
				result.push('\\' + splitterArray[i]);
			} else {
				result.push(splitterArray[i]);
			}
		}

		return result;
	};

	// Склеить массивы с разделителями в один и удалить повторения
	self._distinctRegexParts = function () {
		if (!arguments) {
			return [];
		}

		var result = [];
		for (var i = 0; i < arguments.length; i++) {
			if (!arguments[i].hasOwnProperty('length')) {
				continue;
			}

			for (var j = 0; j < arguments[i].length; j++) {
				if (result.indexOf(arguments[i][j]) === -1) {
					result.push(arguments[i][j]);
				}
			}
		}

		return result;
	};

	// Сделать регулярное выражение 
	// с заданным началом, набором символов по середине и окончанием
	self._createMatchingRegex = function (begin, possibleValues, end) {
		if (!possibleValues || !possibleValues.hasOwnProperty('length')) {
			return new RegExp(begin + end);
		}

		var regexText = begin + '[';
		var escapedSplitters = self._escapeSplitters(possibleValues);

		for (var i = 0; i < escapedSplitters.length; i++) {
			regexText += escapedSplitters[i];
		}

		return new RegExp(regexText + ']' + end);
	};
};